import json
from yac.lib.file import get_file_contents
from yac.lib.template import apply_stemplate
from yac.lib.variables import get_variable

# Return a list of boot script lines, one per script line provided in params.
# List is intended to be incorporated into the UserData portion of a EC2 or ASG cloud formation
# template.
def get_value(params):

	boot_script_list=[]

	services_consumed = get_variable(params, 'services-consumed', [])
	servicefile_path = get_variable(params,"servicefile-path")

	boot_file = get_variable(params,'boot-file',"")

	# get the boot script from the user params
	if boot_file:
		
		boot_script_contents = get_file_contents(boot_file,servicefile_path)

		if boot_script_contents:

			# render template variables into the file contents
			boot_script_contents = apply_stemplate(boot_script_contents, params)

			# split script into lines
			boot_script_lines_list = boot_script_contents.split('\n')

			for i,line in enumerate(boot_script_lines_list):

				if ("{" in line and "}" in line and "Ref" in line):
					# this line contains a cloud formation reference which needs to be broken out
					# i.e. something like ...
					# CLUSTER_NAME={"Ref": "ECS"}

					prefix = line[:line.index('{')]
					reference = line[line.index('{'):line.index('}')+1]
					reference_dict = json.loads(reference)

					boot_script_list = boot_script_list + [prefix,{"Ref": reference_dict["Ref"]}] + ["\n"]

				else:
					boot_script_list = boot_script_list + [line] + ["\n"]	


	else:
		boot_script_list = boot_script_list + ["# No boot script provided. See yac docs for more info.\n"]

	# for debugging
	# pp_script(boot_script_list)

	return boot_script_list

def pp_script(boot_script_list):

	for line in boot_script_list:
		print str(line)




import argparse,sys,os
from sets import Set
import yac.cli.restoredb
import yac.cli.setupdb

from yac.cli.primer import show_primer

def main():

    # first argument is help, show primer
    if (len(sys.argv)==1 or sys.argv[1] == '-h'):

        show_primer(['db','primer'])

    else:

        # strip command from args list
        command = sys.argv[1]
        sys.argv = sys.argv[1:]

        if command == 'setup':

            return yac.cli.setupdb.main()

        elif command == 'restore':

            return yac.cli.restoredb.main()

        else:

            print "Command not supported or known"
            show_primer(['db','primer'])


import os

def get_config_path():

	config_path = os.path.join(os.path.dirname(__file__),'../','config')

	return config_path

def get_root_path():

	root_path = os.path.join(os.path.dirname(__file__),'../')

	return root_path

def get_yac_path():

	yac_path = os.path.join(os.path.dirname(__file__),'../../')

	return yac_path

def get_lib_path():

	lib_path = os.path.join(os.path.dirname(__file__))

	return lib_path	


#!/usr/bin/env python

from docker import Client

# get connection string for connecting to docker on a remote host
# default port is 5555
def get_connection_str( host, port = 5555):	
    return "http://%s:%s" % ( host, port )

# get docker client
def get_docker_client( connection_str = 'unix://var/run/docker.sock'):
    return Client( version='auto', base_url=connection_str )

# find a container by name
# for some reason this method isn't implemented directly in the docker-py module
def find_container_by_name( container_name, connection_str = 'unix://var/run/docker.sock'):

    # get a client connection
    client = get_docker_client( connection_str )

    # get a list of all containers
    containers = client.containers(all=True)

    # initialize null container (returned if we can't find any that match)
    null_container = {}

    # find the container
    for container in containers:

        # gotcha alert, if server has no containers, container['Names'] can include Names=None,
        # so need to be defensive ...
        if (container and 'Names' in container and container['Names']):

            for this_name in container['Names']:

                if (this_name and container_name in this_name):

                    return container

    return null_container

	


#!/usr/bin/env python

import argparse, sys
#from slib.name import get_stack_name
#from slib.stack import get_rds_db_endpoint
#from slib.db import rename_rds_instance, snapshot_exists, restore_intance_from_snapshot
# from slib.db import restore_security_group, delete_instance

def prompt_for_confirmation(msg):

    sys.stdout.write('\r')                            
    sys.stdout.write( '%s%s%s'%('\033[92m',msg, '\033[0m'))
    sys.stdout.flush() 
    raw_input(" ")

def main():

	parser = argparse.ArgumentParser('Restore a DB from an RDS snapshot')

	# required args
	parser.add_argument('app',  help='the app name')
	parser.add_argument('env',  help='the env', 
	                            choices=['dev', 'stage', 'prod','archive'])
	parser.add_argument('snap', help='the name of the RDS snapshot to restore from')
	parser.add_argument('-x','--suffix',  help='app suffix, to support multiple instances of the same app in the same env')
	parser.add_argument('-s','--skipto',  help='skip to this step number')

	# pull out args
	args = parser.parse_args()


#!/usr/bin/env python

import argparse, sys, json, os

from yac.lib.service import validate_service, register_service, clear_service, get_service_by_name
from yac.lib.service import get_service_from_file, publish_service_description, ServiceError

def is_valid_file(parser, arg):
    if not os.path.exists(arg):
        parser.error("The file %s does not exist!" % arg)
    else:
        return arg

def main():

    parser = argparse.ArgumentParser('Install a service in the yac registry (works a bit like an s3 bucket, but w/out the pesky security)')

    parser.add_argument('--validate', 
                        help=('validate the contents of a Servicefile '+
                              '(arg is the path to the Servicefile)'),
                        type=lambda x: is_valid_file(parser, x))
    parser.add_argument('--share', 
                        help=('register a service with the yac service registry so it can be provisioned by others.' +
                        ' (arg is the path to the Servicefile)'),
                        type=lambda x: is_valid_file(parser, x))
    parser.add_argument('--show',  help='show a service (arg is the service key)')    
    parser.add_argument('--clear', help='clear a service from the registry  (arg is the service key)')
    parser.add_argument('--find',  help=('find a service in the registry via a search string '+
                                         '(arg is search string into the registry)')) 

    # pull out args
    args = parser.parse_args()

    if args.validate:

        # pull descriptor from file
        service_descriptor, service_name, servicefile_path = get_service_from_file(args.validate)
        
        # validate the service description
        validation_errors = validate_service(service_descriptor)

        if validation_errors:
            print ("Your service description failed validation checks. Please fix the following errors. %s"%validation_errors)
        else:
            print "Service is ready to be shared!"

    elif args.share:

        # pull descriptor from file
        service_descriptor, service_name, servicefile_path = get_service_from_file(args.share)

        # validate the service description
        validation_errors = validate_service(service_descriptor, servicefile_path)

        if not validation_errors:

            # see if service has already been registered
            service_descriptor_in_registry = get_service_by_name(service_name)

            if not service_descriptor_in_registry:
                challenge = raw_input("Please input a challenge phrase to control updates to your service definition >> ")
            else:
                challenge = raw_input("Please input the challenge phrase associated with this service definition >> ")


            print ("About to register service '%s' with challenge phrase '%s'. ")%(service_name,challenge)
            raw_input("Hit Enter to continue >> ")

            try:
                register_service(service_name, args.share, challenge)

                # publish the service 
                publish_service_description(service_name, args.share)

                print ("Your service has been registered with yac under the key: '%s' and challenge phrase '%s'.\n" +
                       "You will prompted for the challenge phrase if you attempt to update your service.\n" + 
                       "Other users can run your service via '>> yac stack %s ...'")%(service_name,challenge,service_name)

            except ServiceError as e:
                print ("Your service registration failed: %s"%e.msg)

        else:
            print ("Your service description failed validation checks. Please fix the following errors. %s"%validation_errors)
        

    elif args.clear:

        print "Clearing the '%s' service from registry"%(args.clear)
        challenge = raw_input("Please input the challenge phrase associated with this service >> ")        
        raw_input("Hit Enter to continue >> ")

        clear_service(args.clear, challenge)

        # TODO: need some success/fail feedback

    elif args.find:
        print "not yet implemented"

    elif args.show:
        print "not yet implemented"        




#!/usr/bin/env python

from docker import Client
import docker, os, json
from api import get_docker_client, get_connection_str


def remove_image(
            image_tag,
            target_host_ip,
            force_removal = True):

    # get connection string for the docker remote api on the target host
    docker_api_conn_str = get_connection_str( target_host_ip )

    docker_client = get_docker_client( docker_api_conn_str )

    print image_tag + ", " + target_host_ip
    
    # build the image
    print docker_client.remove_image(
                                    image=image_tag,
                                    force=force_removal)


#!/usr/bin/env python

import argparse
from docker import Client
from api import get_docker_client, get_connection_str, find_container_by_name

def get_logs(container_alias, target_host_ip):

	# get connection string for the docker remote api on the target host
	docker_api_conn_str = get_connection_str( target_host_ip )

	docker_client = get_docker_client( docker_api_conn_str )

	# find container
	container = find_container_by_name( container_alias, docker_api_conn_str )

	# get logs from container
	return docker_client.logs(container=container.get('Id'))



import os, mimetypes, shutil
from yac.lib.file import get_file_contents

def apply_stemplate(string_w_variables, template_variables):

    for key in template_variables:
        if 'value' in template_variables[key]:
            variable_value = template_variables[key]['value']
            if (type(variable_value) == str or type(variable_value) == unicode):
                to_replace = "{{%s}}"%key            
                string_w_variables = string_w_variables.replace(to_replace,str(variable_value))

    return string_w_variables

def apply_ftemplate(file_w_variables, template_variables):

    # read file into string
    string_w_variables = get_file_contents(file_w_variables)

    return apply_stemplate(string_w_variables, template_variables)

def apply_templates_in_file(file_w_variables, template_variables, rendered_file_dest="tmp"):

    # get the file type
    file_type = mimetypes.guess_type(file_w_variables)

    # print 'file type for %s: %s'%(file_w_variables,file_type[0])

    # if the file is a text file render any variables in the file contents using the
    # provided template variables
    if ( file_type and 
         len(file_type)>=1 and 
         ( 'text' in file_type[0] or 
           'json' in file_type[0] ) ):

        # read file into string
        file_contents = get_file_contents(file_w_variables)

        # render variables
        rendered_file_contents = apply_stemplate(file_contents, template_variables)    

        # create a 'tmp' directory to hold the rendered file contents
        if not os.path.exists(rendered_file_dest):
            os.makedirs(rendered_file_dest)

        file_name = os.path.basename(file_w_variables)

        rendered_file = os.path.join(rendered_file_dest,file_name)

        # print "rf: %s"%rendered_file

        # write the rendered string into the temp file
        with open(rendered_file, 'w') as outfile:
            outfile.write(rendered_file_contents)
    else:

        # this isn't a text file, so don't attemp to render any variables
        # instead copy from source to destination

        # create a 'tmp' directory to hold the files
        if not os.path.exists(rendered_file_dest):
            os.makedirs(rendered_file_dest)

        file_name = os.path.basename(file_w_variables)

        rendered_file = os.path.join(rendered_file_dest,file_name)

        # print "nrf: %s"%rendered_file

        shutil.copy(file_w_variables, rendered_file)
        

    return rendered_file

def apply_templates_in_dir(source_dir, template_variables, dest_dir="tmp"):

    # get the contents of this directory
    dir_children = os.listdir(source_dir)

    for this_child in dir_children:

        if os.path.isfile(os.path.join(source_dir,this_child)):

            this_file = os.path.join(source_dir,this_child)

            apply_templates_in_file(this_file, template_variables, dest_dir)

        else:

            # destination is relative to the current destination
            new_dest_dir = os.path.join(dest_dir,this_child)

            # source dir is relative to the current source
            new_source_dir = os.path.join(source_dir,this_child)

            apply_templates_in_dir(new_source_dir,template_variables,new_dest_dir)
           


#!/usr/bin/env python

import os, json
import jmespath
from jinja2 import Template

# xform env variables in taskdef format into a format compatible
# with docker api
def env_ec2_to_api(taskdef_envs):

    env_map = {}

    for env in taskdef_envs:
        env_map[env['Name']] = env['Value']

    return env_map

# xform volumes and mountpoints in taskdef format into a format compatible
# with docker api volume maps
def get_volumes_map(taskdef_volumes, taskdef_mountPoints):

    volume_map = {}
    volume_bindings = []
    for volume in taskdef_volumes:

        # example jmespath for 'home' volume is
        # [?sourceVolume=='home'].containerPath
        cp_search_str = "[?SourceVolume=='%s'].ContainerPath"%volume['Name']
        ro_search_str = "[?SourceVolume=='%s'].ReadOnly"%volume['Name']

        bind_path = jmespath.search(cp_search_str,taskdef_mountPoints)
        read_only = jmespath.search(ro_search_str,taskdef_mountPoints)

        # if we found a hit in the mount points, then add this volume to the 
        # volume map
        if (len(bind_path)==1 and len(read_only)==1):

            volume_bindings.append(volume['Host']['SourcePath'])
            
            volume_map[ volume['Host']['SourcePath'] ] = { 
                                    "bind": jmespath.search(cp_search_str,taskdef_mountPoints)[0],
                                    "ro":  jmespath.search(ro_search_str,taskdef_mountPoints)[0]
                                }


    return volume_map, volume_bindings

# xform port mapping in taskdef format into the port binding format compatible
# with docker api 
def get_port_bindings(taskdef_port_mappings):

    port_binding = {}

    for port_mapping in taskdef_port_mappings:
        port_binding[str(port_mapping['ContainerPort'])] = str(port_mapping['HostPort'])

    return port_binding

# render template variable in the container environment variables.
# template_vars contains key/value pairs of template variables
# that may appear in container_envs values
def _render_env_variables(container_envs, template_vars):
    
    for key in container_envs:
        container_envs[key] = _render_string_variables(container_envs[key],
                                                       template_vars)
    return container_envs

def _render_string_variables(string_w_variables, template_vars):

    template = Template(string_w_variables)
    return template.render(template_vars)

# load the config file which defines the container parameters
def load_container_configs(container_alias, 
                           config_file, 
                           env_dynamic = None):

    container_configs = load_configs(config_file)

    if container_alias in container_configs:
        configs = container_configs[container_alias]

    # render dynamic variables
    if env_dynamic:
        configs['envs'] = _render_env_variables(configs['envs'], env_dynamic)
    
    return configs


# taskdef_configs per ECS standard
def get_aliases(taskdef_configs):

    return jmespath.search('ContainerDefinitions[*].name', taskdef_configs)
  
 


#!/usr/bin/env python

from docker import Client
import docker, os, json
from api import get_docker_client

def build_image(
            image_tag,
            sources_path,
            connection_str = "",
            rm = "false"):

    print 'Getting docker client for connection_str=%s' % (connection_str)

    docker_client = get_docker_client( connection_str )

    # build the image
    for line in docker_client.build(tag=image_tag,
                                    path=sources_path,
                                    rm=rm):

        line_json = json.loads(line)

        if 'stream' in line_json:
            print line_json['stream']
        else:
            print line_json

def get_image_name(image, version):

    return "%s:%s"%(image,version)






import argparse,sys,os
import yac.cli.build
import yac.cli.push
import yac.cli.start
#import yac.cli.stop

from yac.cli.primer import show_primer

def main():

    # first argument is help, show primer
    if (len(sys.argv)==1 or sys.argv[1] == '-h'):

        show_primer(['container','primer'])

    else:

        # strip command from args list
        command = sys.argv[1]
        sys.argv = sys.argv[1:]

        if command == 'start':

            return yac.cli.start.main()
        
        #elif command == 'stop':

        #    return yac.cli.stop.main()

        elif command == 'build':

            return yac.cli.build.main()

        elif command == 'push':

            return yac.cli.push.main()

        else:

            print "Command not supported or known"
            show_primer(['container','primer'])


import os, json
from yac.lib.file import register_file, get_file_reg_key, file_in_registry
from yac.lib.registry import clear_entry_w_challenge

# convert references to local files from relative paths to registry paths
def convert_local_files(local_key,dict_contents_str,local_base_path,challenge):

    # convert contents to json
    source_dict = json.loads(dict_contents_str)

    source_dict_updated = find_and_convert_locals(source_dict, local_key, local_base_path, challenge)

    # convert back to a string
    updated_contents_str = json.dumps(source_dict_updated)

    return updated_contents_str

# convert references to local files in the service descriptor, and register
# each source file

def find_and_convert_locals(source_dict, local_key, local_base_path, challenge):

    for key in source_dict.keys():

        if type(source_dict[key])==dict:
            source_dict[key] = find_and_convert_locals(source_dict[key],local_key,local_base_path,challenge)

        elif type(source_dict[key])==list:
            source_dict[key] = find_and_convert_locals_list(source_dict[key],local_key,local_base_path,challenge)

        else:
            source_dict[key] = find_and_convert_locals_leaf(source_dict[key],local_key, local_base_path,challenge)

    return source_dict 

def find_and_convert_locals_list(source_list, local_key, local_base_path, challenge):

    for i, item in enumerate(source_list):
        if type(item)==dict:
            source_list[i] = find_and_convert_locals(item, local_key, local_base_path, challenge)           
        elif type(item)==list:
            source_list[i] = find_and_convert_locals_list(item, local_key, local_base_path, challenge) 
        else:
            source_list[i] = find_and_convert_locals_leaf(item, local_key, local_base_path,challenge)

    return source_list

def find_and_convert_locals_leaf(dict_leaf, local_key, local_base_path, challenge):

    # see if the value is a file, and see if file is referenced relative to the 
    # service file
    if ( os.path.exists(os.path.join(local_base_path,str(dict_leaf))) and
         os.path.isfile(os.path.join(local_base_path,str(dict_leaf))) and
         not os.path.isabs(str(dict_leaf)) ):

        file_path = os.path.join(local_base_path,str(dict_leaf))

        # prefix file path with service key to avoid naming collisions
        file_key = os.path.join(local_key,str(dict_leaf))

        # print "registering %s to %s"%(file_path, file_key)

        register_file(file_key, file_path, challenge)

        file_url = get_file_reg_key(file_key)

        return file_url

    else:

        return dict_leaf

def find_and_delete_remotes(source_dict, challenge):

    for key in source_dict.keys():

        if type(source_dict[key])==dict:
            source_dict[key] = find_and_delete_remotes(source_dict[key],challenge)

        elif type(source_dict[key])==list:
            source_dict[key] = find_and_delete_remotes_list(source_dict[key], challenge)

        else:
            source_dict[key] = find_and_delete_remotes_leaf(source_dict[key],challenge)

    return source_dict 
              
def find_and_delete_remotes_list(source_list, challenge):

    for i, item in enumerate(source_list):
        if type(item)==dict:
            source_list[i] = find_and_delete_remotes(item, challenge)           
        elif type(item)==list:
            source_list[i] = find_and_delete_remotes_list(item, challenge) 
        else:
            source_list[i] = find_and_delete_remotes_leaf(item, challenge)

    return source_list

def find_and_delete_remotes_leaf(item, challenge):

    # see if the value is a file
    if file_in_registry(str(item)):

        clear_entry_w_challenge(str(item), challenge)
        


#!/usr/bin/env python

from docker import Client
import docker, os, json
from api import get_docker_client

def push_image(
            image_tag,
            hub_uname,
            hub_pwd,
            hub_email,
            connection_str = ""):

    print 'Getting docker client for connection_str=%s' % (connection_str)

    docker_client = get_docker_client( connection_str )

    # first login to docker hub
    docker_client.login(hub_uname,hub_pwd,hub_email)

    # push the image
    for line in docker_client.push(image_tag, stream=True):

        line_json = json.loads(line)

        if 'stream' in line_json:
            print line_json['stream']
        else:
            print line_json







import os, sys, json, argparse, getpass, inspect, pprint
from colorama import Fore, Style
from botocore.exceptions import ClientError
from yac.lib.file import FileError
from yac.lib.stack import create_stack, update_stack, cost_stack
from yac.lib.stack import BUILDING, UPDATING, stack_exists, get_stack_templates
from yac.lib.stack import pretty_print_resources, deploy_stack_files
from yac.lib.service import get_service, get_service_parmeters
from yac.lib.naming import set_namer, get_stack_name
from yac.lib.variables import get_variable, set_variable
from yac.lib.intrinsic import apply_fxn, INSTRINSIC_ERROR_KEY
from yac.lib.vpc import get_vpc_prefs

def main():

    parser = argparse.ArgumentParser('Show the status of a yac stack')
    # required args
    parser.add_argument('service_alias', help='name assigned to this service in your cloud'), 

    # optional
    # store_true allows user to not pass a value (default to true, false, etc.)    
    parser.add_argument('-e',
                        '--env',  help='the environment to build stack cluster for', 
                                  choices=  ['dev','stage','prod'],
                                  default= "dev")
    parser.add_argument('--suffix',  help='to create multiple instances of the same app in the same env (e.g. confluence-intranet, confluence-tech)',
                                     default='')                                       

    args = parser.parse_args()

    # get vpc preferences in place
    vpc_prefs = get_vpc_prefs()

    # get the service parameters for use with yac-ref's in service templates
    service_parmeters = get_service_parmeters(args.service_alias, args.env, args.suffix, 
                                              "", {},
                                              "", vpc_prefs)

    # get stack name
    stack_name = get_stack_name(service_parmeters)  

    print(Fore.GREEN)

    # show stack status

    print(Style.RESET_ALL)                      
      



import os, imp, urlparse
from sets import Set
from yac.lib.file import get_file_contents, file_in_registry, file_in_yac_sources, create_customization_file
from yac.lib.paths import get_yac_path, get_lib_path
from yac.lib.variables import get_variable,set_variable
from yac.lib.naming import get_resource_name

INSTINSICS = ['yac-ref', 'yac-join', 'yac-fxn', 'yac-name']
YAC_REF_ERROR = "ref-error"
YAC_FXN_ERROR = "fxn-error"
INSTRINSIC_ERROR_KEY = 'intrinsic-errors'

def apply_fxn(source_dict, params):

    for key in source_dict.keys():

        if type(source_dict[key])==dict:
            source_dict[key] = apply_fxn_dict(source_dict[key], params)

        elif type(source_dict[key])==list:
            source_dict[key] = apply_fxn_list(source_dict[key], params)

        else:
            source_dict[key] = apply_fxn_leaf(key,source_dict, params)

    return source_dict 

def apply_fxn_dict(source_dict, params):

    sub_keys = source_dict.keys()
    if len(Set(sub_keys) & Set(INSTINSICS))==1:
        # treat this as a leaf
        source_dict = apply_fxn_leaf(sub_keys[0],source_dict, params)
    else:
        source_dict = apply_fxn(source_dict,params)

    return source_dict

def apply_fxn_list(source_list, params):

    for i, item in enumerate(source_list):
        if type(item)==dict:
            source_list[i] = apply_fxn_dict(item, params)           
        elif type(item)==list:
            source_list[i] = apply_fxn_list(item, params) 
        else:
            source_list[i] = item
    return source_list
              
def apply_fxn_leaf(key, source_dict, params):

    # see if any of the values have intrinsics
    if key == 'yac-ref':

        # Pull referenced value from the params. Default to a string
        # containing an error message in case the reference does not have
        # a corresponding value.

        setpoint = get_variable(params,source_dict[key],"M.I.A.")
        if setpoint=="M.I.A.":
            setpoint = '%s: %s'%(YAC_REF_ERROR,source_dict[key])
            error_list = get_variable(params,INSTRINSIC_ERROR_KEY,[])
            set_variable(params,INSTRINSIC_ERROR_KEY,error_list+[setpoint])

        return setpoint

    elif key == 'yac-join':

        delimiters = source_dict[key][0]
        name_parts = source_dict[key][1]

        # apply any intrinsics in list
        filled_parts = apply_fxn_list(name_parts, params)

        # get rid of empty strings before joining with delimitter
        filled_parts = filter(None,filled_parts)

        return delimiters.join(filled_parts)

    elif key == 'yac-fxn':

        # this value should be filled by custom function supplied by service
        fxn_script = source_dict[key]
        return apply_custom_fxn(fxn_script, params)

    elif key == 'yac-name':

        # get the name for this resource
        resource = source_dict[key]
        return get_resource_name(params, resource)

    elif key == 'yac-include':
        
        include_file = source_dict[key]

        print "to include: %s"%include_file

        # get dict from file
        # included_dict = get_file_contents(include_file)

        # print included_dict
        # render any params into the dictionary
        #return apply_fxn(included_dict, params)
        return {}

    else:

        return source_dict[key]

def apply_custom_fxn(script_path_arg, params):

    # get the python file that will be used to build this param value
    script_path = get_params_script_path(script_path_arg, params)

    return_val = ""

    if (script_path and os.path.exists(script_path)):

        # module_name = 'yac.lib.customizations.%s.params'%app_alias
        module_name = 'yac.lib.customizations'
        script_module = imp.load_source(module_name,script_path)

        # call the get_value fxn in the script
        return_val = script_module.get_value(params)

    else:
        setpoint = '%s: %s'%(YAC_FXN_ERROR,script_path)
        error_list = get_variable(params,INSTRINSIC_ERROR_KEY,[])
        set_variable(params,INSTRINSIC_ERROR_KEY,error_list+[setpoint])
        
    return return_val 

def get_params_script_path(script_path_arg, params):

    # if the path input is an yac url, download from the registry to a local
    # file
    if file_in_registry(script_path_arg):

        script_file_path = create_customization_file(script_path_arg)
    
    elif file_in_yac_sources(script_path_arg):
        
        script_file_path = os.path.join(get_yac_path(),script_path_arg)

    else:
        # assume the script path is local, and that path is relative to the location
        # of the service file (just like Dockerfile!)
        servicefile_path = get_variable(params,"servicefile-path")
        script_file_path = os.path.join(servicefile_path,script_path_arg)

        print "assuming %s"%script_path_arg

    return script_file_path



from sets import Set

def validate_dictionary(dictionary, supported_keys, required_fields):

    validation_errors = {}

    missing_fields = find_missing_fields(dictionary, required_fields)

    unsupported_keys = find_unsupported_keys(dictionary, supported_keys)

    if missing_fields:
        validation_errors["missing-fields"] = missing_fields
    if unsupported_keys:
        validation_errors["unsupported-keys"] = unsupported_keys

    return validation_errors

def find_unsupported_keys(dictionary, supported_keys):

    key_ustring_list = dictionary.keys()

    key_string_list = []
    # convert all keys from unicode to normal string
    for key_ustring in key_ustring_list:
        key_string_list = key_string_list + [str(key_ustring)]

    unsupported_keys = list( Set(key_string_list) - Set(supported_keys))

    return unsupported_keys

# TODO. I tried to do this with with jmespath, which is way more elegant,
# but it doesn't seem to support hyphens in json keys. lame
def find_missing_fields(dictionary, required_fields):

    missing_fields = []

    for i,field_path in enumerate(required_fields):

        path_parts = field_path.split(".")

        path_len = len(path_parts)

        if (path_len == 1 and 
            path_parts[0] not in dictionary):

            missing_fields = missing_fields + [required_fields[i]]
        
        elif ( path_len == 2 and
               ( (path_parts[0] not in dictionary) or 
                 (path_parts[1] not in dictionary[path_parts[0]])
               )
             ):

            missing_fields = missing_fields + [required_fields[i]]

        elif ( path_len == 3 and
               ( (path_parts[0] not in dictionary) or 
                 (path_parts[1] not in dictionary[path_parts[0]]) or
                 (path_parts[2] not in dictionary[path_parts[0]][path_parts[1]] )
               )
             ):

            missing_fields = missing_fields + [required_fields[i]]

        elif ( path_len == 4 and
               ( (path_parts[0] not in dictionary) or 
                 (path_parts[1] not in dictionary[path_parts[0]]) or
                 (path_parts[2] not in dictionary[path_parts[0]][path_parts[1]]) or
                 (path_parts[3] not in dictionary[path_parts[0]][path_parts[1]][path_parts[2]])
               )
             ):

            missing_fields = missing_fields + [required_fields[i]]            

    return missing_fields



import os, json, jmespath, boto3
from botocore.exceptions import ClientError
from yac.lib.paths import get_config_path
from yac.lib.registry import set_remote_string_w_challenge, get_remote_value, get_registry_keys
from yac.lib.registry import set_local_value, get_local_value, clear_entry_w_challenge
from yac.lib.validator import validate_dictionary
from yac.lib.file_converter import convert_local_files, find_and_delete_remotes
from yac.lib.variables import get_variable, set_variable
from yac.lib.file import get_file_contents, create_customization_file, get_file_reg_key

YAC_VPC_SUFFIX="-vpc"

REQUIRED_FIELDS=["prefs-name.value",
                 "vpc-params"]

SUPPORTED_KEYS=[ "prefs-repo",
                 "prefs-name",
                 "vpc-inputs",
                 "vpc-params",
                 "resource-namer",
                 "service-inputs",
                 "service-templates"]

def validate_vpc_prefs(vpc_prefs):

    return validate_dictionary(vpc_prefs, SUPPORTED_KEYS, REQUIRED_FIELDS)

def get_vpc_prefs():

    vpc_preferences = get_local_value('vpc_preferences')

    return vpc_preferences

def set_vpc_prefs(vpc_prefs):

    if vpc_prefs:

        # save vpc preferences to local db
        set_local_value('vpc_preferences',vpc_prefs)

def clear_vpc_prefs():

    set_local_value('vpc_preferences',{})

# register namer into yac registry
def register_vpc_prefs(vpc_prefs_key, vpc_prefs_file, challenge):

    with open(vpc_prefs_file) as vpc_prefs_file_fp:

        vpc_preferences_str = vpc_prefs_file_fp.read()

        vpc_prefs_file_path = os.path.dirname(vpc_prefs_file)

        updated_vpc_preferences_str = convert_local_files(vpc_prefs_key,
                                          vpc_preferences_str,
                                          vpc_prefs_file_path,
                                          challenge)

        # add the suffix to distinguish this key as a vpc preferences key
        vpc_prefs_registry_key = vpc_prefs_key + YAC_VPC_SUFFIX

        # set the vpc defs in the registry
        set_remote_string_w_challenge(vpc_prefs_registry_key, updated_vpc_preferences_str, challenge)

def clear_vpc_prefs_from_registry(vpc_prefs_name, challenge):

    # if vpc preferences are in fact registered
    vpc_prefs = get_vpc_prefs_from_registry(vpc_prefs_name)
    
    if vpc_prefs:

        # add the suffix to distinguish this key as a vpc preferences key
        vpc_prefs_registry_key = vpc_prefs_name + YAC_VPC_SUFFIX

        # clear the vpc defs from the registry
        clear_entry_w_challenge(vpc_prefs_registry_key, challenge)

        # clear any files referenced 
        find_and_delete_remotes(vpc_prefs, challenge) 

def get_vpc_prefs_from_file(prefs_file_path):

    vpc_prefs = {}
    vpc_prefs_name = ""

    with open(prefs_file_path) as vpc_prefs_file_fp:

        vpc_prefs = json.load(vpc_prefs_file_fp)

    return vpc_prefs

def get_vpc_prefs_from_registry(vpc_def_registry_key):

    vpc_prefs = {}

    if vpc_def_registry_key:

        # get vpc defs from registry
        vpc_prefs_str = get_remote_value('%s-vpc'%vpc_def_registry_key)

        if vpc_prefs_str:
            # convert back to dictionary
            vpc_prefs = json.loads(vpc_prefs_str)

    return vpc_prefs

def get_all_vpc_def_keys():

    vpc_prefs = []

    # get all registry keys
    registry_keys = get_registry_keys()

    # find all keys with -vpc suffix
    for key in registry_keys:

        if '-vpc' in key:
            # remove the naming part
            vpc_prefs = vpc_prefs + [key.replace('-vpc','')]

    return vpc_prefs  

# get vpc preferences from a local file
def get_vpc_prefs_from_local_file(vpc_prefs_file):

    vpc_prefs = get_vpc_prefs_from_file(vpc_prefs_file)

    vpc_prefs_name = get_variable(vpc_prefs,'prefs-name',"")

    prefs_path = os.path.dirname(vpc_prefs_file)

    # If the vpc prefs references a vpc-inputs script, then we need to save it
    # as a yac customization so it can be executed against any Scriptfile.

    # see if an inputs script is specified
    vpc_inputs_script_path = get_variable(vpc_prefs,'service-inputs',"")
    
    if vpc_inputs_script_path:

        # use the preferences name for the inputs file namespace
        yac_file_key = get_file_reg_key(vpc_inputs_script_path,vpc_prefs_name)

        prefs_file_contents = get_file_contents(os.path.join(prefs_path,vpc_inputs_script_path))

        # save the file under yac/lib/customizations
        inputs_file_yac_path = create_customization_file(yac_file_key,prefs_file_contents)

        # save the  new path back to the preferences
        vpc_prefs['service-inputs']['value'] = inputs_file_yac_path

    # If the vpc prefs references a namer script, then we need to save it
    # as a yac customization so it can be executed against any Scriptfile.

    # see if a namer script is specified
    vpc_namer = get_variable(vpc_prefs,'resource-namer',"")

    if vpc_namer:

        # use the preferences name for the namer file namespace
        yac_file_key = get_file_reg_key(vpc_namer,vpc_prefs_name)

        namer_contents = get_file_contents(os.path.join(prefs_path,vpc_namer))

        # save the file as a yac-local customization
        namer_file_yac_path = create_customization_file(yac_file_key,namer_contents)

        # change file reference so that instrincs recognizes it as part of the yac sources
        set_variable(vpc_prefs,'resource-namer', namer_file_yac_path)

    # If the vpc prefs references service templates, then we need to save them
    # under yac customization so they can be accessed locally.

    if 'service-templates' in vpc_prefs:

        for key in vpc_prefs['service-templates'].keys():

            template_file_name = vpc_prefs['service-templates'][key]

            # use the preferences name for the inputs file namespace
            yac_file_key = get_file_reg_key(template_file_name,vpc_prefs_name)

            template_file_contents = get_file_contents(os.path.join(prefs_path,template_file_name))

            # save the file under yac/lib/customizations
            inputs_file_yac_path = create_customization_file(yac_file_key,template_file_contents)

            # save the file as a yac-local customization
            vpc_prefs['service-templates'][key] = inputs_file_yac_path

    return vpc_prefs 

# get the subnets to use for a given vpc and a given set of 
# availability zones
def set_dmz_subnets( vpc_id ):

    # get the env
    env  = get_variable(params,'env')

    # get the name of the stack
    stack_name = get_stack_name(params)

    subnet_ids=[]

    if not stack_exists(stack_name):

        # see if the subnets are already defined (usually via the vpc prefs)
        dmz_subnets_prefs = get_variable("dmz-subnet-ids")

        if not dmz_subnets_prefs:

            # use the wizard to prompt user for subnets
            subnet_ids = subnets_wizard()

        else:
            subnet_ids = vpcs_per_prefs

    else:

        # determine the existing private subnets in use for this service
        subnet_ids = get_stack_private_subnets(params)

    return vpc

# get the availability zones to use for a given environment
def _get_availability_zones(params):

    azs = {}

    # get the env
    env  = get_variable(params,'env')

    # get the name of the stack
    stack_name = get_stack_name(params)

    if not stack_exists(stack_name):

        # get availability zones from params
        availabilty_zones = get_variable(params,"availability-zones",[])

        # see if the vpc is already defined (usually via the vpc prefs)
        azs_per_prefs = get_variable(params,'vpc-id')

        if not vpcs_per_prefs:

            # use the default vpc for this env, per vpc_mapping
            azs = get_vpc_wizard()

        else:
            azs = vpcs_per_prefs

    else:

        # determine the existing VPC in use for this service
        azs = get_stack_vpc(params)

    return azs    




import argparse, json, os
from yac.lib.vpc import validate_vpc_prefs, get_vpc_prefs, register_vpc_prefs
from yac.lib.vpc import set_vpc_prefs, get_vpc_prefs_from_registry, get_vpc_prefs_from_file
from yac.lib.vpc import get_all_vpc_def_keys, clear_vpc_prefs, clear_vpc_prefs_from_registry
from yac.lib.vpc import get_vpc_prefs_from_local_file
from yac.lib.paths import get_config_path
from yac.lib.variables import get_variable, set_variable
from yac.lib.registry import set_private_registry, clear_private_registry

def is_valid_file(parser, arg):
    if not os.path.exists(arg):
        parser.error("The file %s does not exist!" % arg)
    else:
        return arg

def get_prefs_from_set(set_arg):

    validation_errors = {}

    if os.path.exists(set_arg):

        # get the preferences from file, and ensure any vpc-inputs have a 
        # absolute path
        vpc_prefs = get_vpc_prefs_from_local_file(set_arg)

        # validate the preferences
        validation_errors = validate_vpc_prefs(vpc_prefs)

    else:
        # treat the arg as the preferences key in the registry, and attempt to retrieve            
        vpc_prefs = get_vpc_prefs_from_registry(set_arg)

    return vpc_prefs, validation_errors

def main():

    parser = argparse.ArgumentParser('Manage my yac preferences')

    # optional args
    parser.add_argument('--registry',  help='yac will prompt your for the connection params for a private registry (or clear current private registry setting)',
                                       action='store_true')
    parser.add_argument('--current',  help='show the preferences currently configured',
                                      action='store_true')
    parser.add_argument('--show',  help='show a set of vpc preferences from the registry (arg is preferences name in registry)')    
    parser.add_argument('--set',   help=('set the vpc preferences to use when building stacks. ' +
                                         '(arg is prefs name for registry lookup or a path to vpc preferences file)'))
    parser.add_argument('--list',  help='list keys of all vpc preferences in the registry',
                                   action='store_true') 
    parser.add_argument('--share', help=('publish vpc preferences to the registry so they can be easily shared.'+
                                         ' arg is path to file containing the preferences'),
                                      type=lambda x: is_valid_file(parser, x))
    parser.add_argument('--clear', help='clear a vpc preference from registry (arg of "local" will clear vpc preferences currently in use locally')    
    
    args = parser.parse_args()       

    if args.registry:

        redis_host = raw_input("Please input the name of the Redis host you would like to use for your yac private registry (press Enter to clear current setting) >> ")

        if redis_host:
            redis_port = raw_input("Please input the port the Redis host listens on >> ")
            private_registry = {"host": redis_host, "port": redis_port}

            set_private_registry(private_registry)
        else:
            print "Clearing private registry preferences currently in place"
            raw_input("Hit Enter to continue >> ")
            clear_private_registry()

    if args.current:

        vpc_prefs = get_vpc_prefs()

        print json.dumps(vpc_prefs, indent=4)

    if args.show:

        vpc_prefs = get_vpc_prefs_from_registry(args.show)

        print json.dumps(vpc_prefs,indent=4)        

    if args.share:

        # pull preferences from file
        vpc_prefs = get_vpc_prefs_from_file(args.share)

        vpc_prefs_name = get_variable(vpc_prefs,"prefs-name","")

        # see if these prefs have already been registered
        vpc_prefs_in_registry = get_vpc_prefs_from_registry(vpc_prefs_name)

        if not vpc_prefs_in_registry:
            challenge = raw_input("Please input a challenge phrase to control updates to your vpc preferences >> ")
        else:
            challenge = raw_input("Please input the challenge phrase associated with this vpc preference >> ")

        print ("About to register vpc definition '%s' with challenge phrase '%s'. ")%(vpc_prefs_name,challenge)
        raw_input("Hit Enter to continue >> ")

        register_vpc_prefs(vpc_prefs_name, args.share, challenge)

        print ("Your vpc preferences have been registered with yac under the key: '%s'.\n" +
               "Other users can configure yac with these vpc prefs via '>> yac prefs --set=%s'")%(vpc_prefs_name,vpc_prefs_name)      

    if args.set:

        vpc_prefs, validation_errors = get_prefs_from_set(args.set)

        if (not validation_errors and vpc_prefs):

            print "Setting vpc preferences locally per '%s'."%args.set
            raw_input("Hit Enter to continue >> ")

            set_vpc_prefs(vpc_prefs)

        elif validation_errors:
            print ("Your vpc preferences failed validation checks. Please fix the following errors. %s"%validation_errors)
        else:
            print ("The preferences key input doesn't exist in the configured registry.\n" +
                   "Use >> yac prefs --list to see valid keys.")

    if args.list:

        print "The following vpc preferences are currently available in the registry"
        print get_all_vpc_def_keys() 

    if args.clear:

        if args.clear == 'local':
            print "Clearing vpc preferences currently in place"
            raw_input("Hit Enter to continue >> ")
            clear_vpc_prefs()
        else:            
            # make sure key is legit
            vpc_prefs = get_vpc_prefs_from_registry(args.clear)
            if vpc_prefs:
                challenge = raw_input("Please input the challenge phrase associated with this vpc preference >> ")
                print "Clearing the '%s' vpc preferences from registry"%args.clear
                raw_input("Hit Enter to continue >> ")
                clear_vpc_prefs_from_registry(args.clear, challenge)
            else:
                print "VPC preferences with the '%s' key do not exist in the registry"%args.clear                




import os, imp, urlparse

from yac.lib.paths import get_config_path, get_lib_path
from yac.lib.registry import set_remote_string_w_challenge, get_remote_value, get_registry_keys
from yac.lib.registry import set_local_value, get_local_value, delete_local_value
from yac.lib.variables import get_variable
from yac.lib.file import get_file_contents, create_customization_file, file_in_registry, get_file_reg_key

# the three functions that all services leverage.
# services can either use the defaults, or provide their
# own versions

def get_stack_name( params ):

    return get_namer_module().get_stack_name( params )

def get_resource_name( params , resource):

    return get_namer_module().get_resource_name( params , resource)

def get_namer_module():

    return imp.load_source('yac.lib.naming',get_namer())

def get_namer():

    yac_namer = get_local_value('yac_namer')

    if not yac_namer:

        # load default namer
        yac_namer = os.path.join( get_lib_path(),'naming_default.py')

    return yac_namer

def set_namer(service_descriptor, servicefile_path, vpc_prefs):

    # if a namer is specified as part of the servicefile
    if get_variable(service_descriptor, "resource-namer",""):

        service_namer_path = get_variable(service_descriptor, "resource-namer","")

        # if the file is not in the registry
        if not file_in_registry(service_namer_path):

            # assume a local path relative to the servicefile path
            service_namer_abs_path = os.path.join(servicefile_path,service_namer_path)

            if os.path.exists(service_namer_abs_path):

                # get file contents
                namer_code = get_file_contents(service_namer_abs_path)

                # associate the file with the the service name
                file_namespace = get_variable(service_descriptor,'service-name',"")
                yac_file_key = get_file_reg_key(service_namer_path,file_namespace)

                # save as a yac-local customization
                service_namer_path = create_customization_file(yac_file_key,namer_code)

        elif file_in_registry(service_namer_path):

            # save the file under yac/lib/customizations, keyed by service name
            service_namer_path = create_customization_file(service_namer_path)

    # if a namer is specified as part of the vpc preferences
    elif get_variable(vpc_prefs, "resource-namer",""):

        # Use the namer specified in the VPC preferences
        # If namer comes via the vpc preferences, it will already have been converted into a customization
        service_namer_path = get_variable(vpc_prefs, "resource-namer","") 

    else:

        # No custom namer was provided by the user.
        # Use the default namer
        service_namer_path = os.path.join( get_lib_path(),'naming_default.py')

    # save namer path to local variable
    set_local_value('yac_namer', service_namer_path)



# yac variables are stored as:
# 
#{
#    "variable-name": {
#        "comment": "variable explanation",
#        "value":   "variable value"
#    }   
#}

def get_variable(params, variable_name, default_value="", value_key='value'):

    if variable_name in params and value_key in params[variable_name]:
        return params[variable_name][value_key]
    else:
        return default_value

def set_variable(params, variable_name, value, comment="", value_key='value'):

    params[variable_name] = {value_key : value, 'comment': comment}
    


import redis   # for remote registry
import shelve  # for local registry
import os
from yac.lib.paths import get_config_path

# the url and port for the yac registry in redis
PUBLIC_REGISTRY = {"host": "pub-redis-14705.us-east-1-3.4.ec2.garantiadata.com",
                    "port": 14705 }
PRIVATE_REGISTRY_DB_KEY = 'private_registry'

class RegError():
    def __init__(self, msg):
        self.msg = msg

def _get_registry():

    # if a private registry has been configured
    private_registry = get_local_value(PRIVATE_REGISTRY_DB_KEY)

    if private_registry:
        registry = redis.Redis(host=private_registry['host'], port=private_registry['port'])
    else:
        registry = redis.Redis(host=PUBLIC_REGISTRY['host'], port=PUBLIC_REGISTRY['port'])

    return registry

def get_registry_keys():

    registry = _get_registry()

    return registry.keys()

def set_remote_string_w_challenge(key_name, string_value, challenge_phrase=""):

    original_challenge = ""

    already_registered = get_remote_value(key_name)

    # see if value is already set
    if already_registered:

        # get the original challenge phrase
        original_challenge = _get_remote_challenge(key_name)

    if not challenge_phrase:
        # prompt use for a phrase
        challenge_phrase = raw_input("Enter your challenge phrase >> ")

    # set the value in the registry if 
    # 1) value is not already registered, or 
    # 2) the challenge phrase matches
    if (not already_registered or challenge_phrase == original_challenge):

        value = {"challenge": challenge_phrase, "value": string_value}
        registry = _get_registry()

        registry.hmset(key_name, value)
    else:
        print key_name, original_challenge

        raise RegError("challenge phrase mismatch") 

def get_remote_value(key_name):

    registry = _get_registry()

    entry = registry.hgetall(key_name)

    if (entry and type(entry)==dict):       
        return entry['value']    
    elif entry:
        return entry
    else:
        return ""

def _get_remote_challenge(key_name):

    registry = _get_registry()

    entry = registry.hgetall(key_name)

    if entry:        
        return entry['challenge']    
    else:
        return ""        

def clear_entry_w_challenge(key_name, challenge_phrase=""):

    original_challenge = ""

    already_registered = get_remote_value(key_name)

    # see if value is already set
    if already_registered:

        # get the original challenge phrase
        original_challenge = _get_remote_challenge(key_name)

    if not challenge_phrase:
        # prompt use for a phrase
        challenge_phrase = raw_input("Enter your challenge phrase >> ")

    # clear the value in the registry if 
    # 1) value is not already registered, or 
    # 2) the challenge phrase matches
    if (not already_registered or challenge_phrase == original_challenge):

        _delete_registry_value(key_name)
    else:
        raise RegError("challenge phrase mismatch")    

def _delete_registry_value(key_name):

    print 'clearing %s'%key_name
    
    registry = _get_registry()

    registry.delete(key_name)

def set_local_value(key_name, value):

    # save value to shelve db
    yac_db = _get_local_db()

    yac_db[key_name] = value

def get_local_value(key_name):  

    local_db = _get_local_db()

    if key_name in local_db:
        return local_db[key_name]
    else:
        return ""

def delete_local_value(key_name):   

    local_db = _get_local_db()

    if key_name in local_db:
        local_db.pop(key_name)

def get_local_keys():

    local_db = _get_local_db()

    return local_db.keys()        

def set_private_registry(private_registry_desc):

    if private_registry_desc:

        # save vpc preferences to local db
        set_local_value(PRIVATE_REGISTRY_DB_KEY,private_registry_desc)

def clear_private_registry():

    set_local_value(PRIVATE_REGISTRY_DB_KEY,{})

def _get_local_db():

    yac_db_path = os.path.join( get_config_path(),'db','yac_db')

    local_db = shelve.open(yac_db_path)

    return local_db



#!/usr/bin/env python
import os, json, urlparse, boto3, subprocess, shutil, jmespath
from botocore.exceptions import ClientError
from sets import Set
from yac.lib.file import FileError, file_in_registry, get_file_contents, localize_file
from yac.lib.template import apply_templates_in_file, apply_templates_in_dir
from yac.lib.paths import get_config_path
from yac.lib.intrinsic import apply_fxn
from yac.lib.variables import get_variable, set_variable
from yac.lib.naming import get_stack_name

UPDATING = "Updating"
BUILDING = "Building"

STACK_STATES = ['CREATE_IN_PROGRESS','CREATE_FAILED','CREATE_COMPLETE','ROLLBACK_IN_PROGRESS',
                'ROLLBACK_FAILED','ROLLBACK_COMPLETE','DELETE_IN_PROGRESS','DELETE_FAILED',
                'DELETE_COMPLETE','UPDATE_IN_PROGRESS','UPDATE_COMPLETE_CLEANUP_IN_PROGRESS',
                'UPDATE_COMPLETE','UPDATE_ROLLBACK_IN_PROGRESS','UPDATE_ROLLBACK_FAILED',
                'UPDATE_ROLLBACK_COMPLETE_CLEANUP_IN_PROGRESS','UPDATE_ROLLBACK_COMPLETE']

NEWLY_CREATED_STATE = 'CREATE_COMPLETE'

UPDATABLE_STATES = ['CREATE_COMPLETE','UPDATE_COMPLETE','UPDATE_ROLLBACK_COMPLETE']

UNKNOWN_STATE = "unknown"

NON_EXISTANT = "non-existant"

# determine cost of a new stack
def cost_stack( template_string="", stack_params = None ):

    cost_response = ""

    client = boto3.client('cloudformation')

    try:

        response = client.estimate_template_cost(TemplateBody=template_string,
                                                 Parameters = stack_params)

        cost_response = str(response['Url'])

    except ClientError as e:
        
        cost_response = 'Stack costing failed: %s'%str(e)

    return cost_response

def create_stack( stack_name , 
                  template_string="", 
                  stack_params = [],
                  stack_tags = []):

    client = boto3.client('cloudformation')

    response = client.create_stack(StackName=stack_name,
                                 TemplateBody=template_string,
                                 Parameters = stack_params,
                                 Tags=stack_tags,
                                 Capabilities=['CAPABILITY_IAM'])

    print 'Stack creation in progress - use AWS console to watch construction and/or see errors'

    return response


def update_stack( stack_name , 
                  template_string="", 
                  stack_params = [],
                  stack_tags = []):

    client = boto3.client('cloudformation')

    response = client.update_stack(StackName=stack_name,
                                   TemplateBody=template_string,
                                   Parameters = stack_params,
                                   Capabilities=['CAPABILITY_IAM'])

    print 'Stack update in progress - use AWS console to watch updates and/or see errors'

    return response       

def get_stack_state( stack_name ):

    client = boto3.client('cloudformation')

    if stack_exists( stack_name ):

        response = client.describe_stacks(StackName=stack_name)

        states = jmespath.search("Stacks[*].StackStatus",response)

        if len(states) == 1:
            state = states[0]
        else:
            state = UNKNOWN

    else:
        state = NON_EXISTANT

    return state 

def get_stack_templates(service_descriptor, service_parmeters):

    service_template = {}

    # get the template from the service_descriptor
    if 'stack-template' in service_descriptor:

        # get the template from the descriptor
        service_template_with_refs = service_descriptor['stack-template']

        # get any resources that are to be excluded based on user input
        resource_exclusions = get_variable(service_parmeters,'resource-exclusions',[])

        for resource in resource_exclusions:
            if resource in service_template_with_refs['Resources']:
                print "removing the %s resource per requested exclusion ..."%resource
                service_template_with_refs['Resources'].pop(resource) 

        # render intrinsics
        service_template = apply_fxn(service_template_with_refs, service_parmeters)

    return service_template

def get_stack_templates_old(service_descriptor, vpc_prefs, service_parmeters):

    # get the services this service consumes
    services_consumed = get_variable(service_descriptor,'services-consumed',[])

    # get the template files to use based on services specified
    template_files = _get_stock_template_files(services_consumed, vpc_prefs) 

    # combine all resources into a single stack dictionary
    stack_template = {}

    for key in template_files.keys():

        print "merging in the cf template in %s for the %s service ..."%(template_files[key],key)
        
        # pull template from file into a dictionary while also rendering any service 
        # parmeters into the dictionary
        this_resource_template = get_stackdef_from_file(template_files[key],
                                                        service_parmeters)

        # merge this resource into the stack definition
        stack_template = update_template(stack_template,this_resource_template)


    # next add any resources or resource changes specified in the service_descriptor
    if 'stack-template' in service_descriptor:

        # first render intrinsics
        service_template = apply_fxn(service_descriptor['stack-template'], service_parmeters)

        # merge these resources into the stack definition
        stack_template = update_template(stack_template,service_template)


    return stack_template

def _get_stock_template_files(services_consumed, vpc_prefs):

    stock_templates = {}
    vpc_prefs_template_map = {}

    # get template map provided in the vpc preferences
    if 'service-templates' in vpc_prefs:
        vpc_prefs_template_map = vpc_prefs['service-templates']

    # default templates supported
    template_map = {
        "asg":   "yac-asg-stack.json",
        "rds":   "yac-db-stack.json",
        "e-elb": "yac-eelb-stack.json",
        "i-elb": "yac-ielb-stack.json",
        "efs":   "yac-efs-stack.json"
    }

    # override the default template map per vpc preferences
    template_map.update(vpc_prefs_template_map)

    config_path = os.path.join(get_config_path(),'stack')

    for resource_key in services_consumed:

        if not file_in_registry(template_map[resource_key]):
            # use the local file
            stock_templates[resource_key] = os.path.join(config_path,template_map[resource_key])
        else:
            # use the file in the registry
            stock_templates[resource_key] = template_map[resource_key]

    return stock_templates

# Inputs are two cloud formation templates.
# Merge the to_add_template into the base_template
def update_template(base_template, to_add_template):

    if 'Description' in to_add_template:
        # add a stack-specific description
        base_template['Description'] = to_add_template['Description']

    if 'Conditions' in to_add_template:

        # first make sure Parameters is present in base template
        if 'Conditions' not in base_template:
            base_template['Conditions'] = {}

        for key in to_add_template['Conditions'].keys():

            base_template['Conditions'][key] = to_add_template['Conditions'][key]

    if 'Parameters' in to_add_template:

        # first make sure Parameters is present in base template
        if 'Parameters' not in base_template:
            base_template['Parameters'] = {}

        for key in to_add_template['Parameters'].keys():

            base_template['Parameters'][key] = to_add_template['Parameters'][key]

    if 'Mappings' in to_add_template:

        # first make sure Resources is present in base template
        if 'Mappings' not in base_template:
            base_template['Mappings'] = {}

        for key in to_add_template['Mappings'].keys():

            base_template['Mappings'][key] = to_add_template['Mappings'][key]

    if 'Resources' in to_add_template:

        # first make sure Resources is present in base template
        if 'Resources' not in base_template:
            base_template['Resources'] = {}

        for key in to_add_template['Resources'].keys():

            base_template['Resources'][key] = to_add_template['Resources'][key]            

    if 'ResourceTweaks' in to_add_template:

        # 'tweak' resources in the standard stack
        for sg_key in to_add_template['ResourceTweaks'].keys():

            # assume Security group changes are additions
            if to_add_template['ResourceTweaks'][sg_key]['Type'] == "AWS::EC2::SecurityGroup":

                for second_tier_key in to_add_template['ResourceTweaks'][sg_key]['Properties'].keys():

                    if second_tier_key == "SecurityGroupEgress":

                        # add special egress rule
                        base_template = _append_sg_rule(base_template, 
                                                        to_add_template,
                                                        sg_key,
                                                        "SecurityGroupEgress")
                
                    elif second_tier_key == "SecurityGroupIngress":

                        # add special ingress rule
                        base_template = _append_sg_rule(base_template, 
                                                        to_add_template,
                                                        sg_key,
                                                        "SecurityGroupIngress")

            # add property changes
            else:

                for second_tier_key in to_add_template['ResourceTweaks'][sg_key]['Properties'].keys():

                    base_template['Resources'][sg_key]['Properties'][second_tier_key] = to_add_template['ResourceTweaks'][sg_key]['Properties'][second_tier_key]                     

    return base_template

def _append_sg_rule(base_template, to_add_template, sg_key, rule_type):

    for egress_rule in to_add_template['ResourceTweaks'][sg_key]['Properties'][rule_type]:

        # add and rules to base template
        base_template['Resources'][sg_key]['Properties'][rule_type].append(egress_rule)

    return base_template    

# returns stack template as a dictionary.
# render any yac intrinsics present in the dictionary
def get_stackdef_from_file(template_path, service_parmeters={}):

    stack_definitions = {}

    stack_defs_str = get_file_contents(template_path)

    if stack_defs_str:
        
        stack_definitions = json.loads(stack_defs_str)

        # render yac intrinsics in the stack definition
        stack_definitions = apply_fxn(stack_definitions, service_parmeters) 

    return stack_definitions

def stack_exists(stack_name):

    client = boto3.client('cloudformation')

    try:
        response = client.describe_stacks(StackName=stack_name)
        stack_count = len(response['Stacks'])
        return stack_count>0
    except:
        return False

def deploy_stack_files(service_descriptor, service_parmeters, servicefile_path): 

    # deploy any individual files that the service needs before booting
    # (default to empty array in case variable is not defined)
    files_for_boot = get_variable(service_descriptor, 'deploy-for-boot', [], 'files')

    # render service parmeters in the files then deploy 'em
    _load_files(files_for_boot, service_parmeters, servicefile_path)

    # deploy any directories that the service needs before booting
    dirs_for_boot = get_variable(service_descriptor, 'deploy-for-boot', [], 'directories')

    # render service parmeters in the files contained in the directories then deploy
    # the directory structure to destination
    _load_dirs(dirs_for_boot, service_parmeters)


# Render service parmeters into file body, then load files to destination
# Only s3 destinations are currently supported.
def _load_files(files, service_parmeters, servicefile_path):

    # assume the file path is relative to the location
    # of the service descriptor file (just like Dockerfile!)
    servicefile_path = get_variable(service_parmeters,"servicefile-path")

    for this_ifile in files:

        # render intrinsics in the file dictionary
        this_file = apply_fxn(this_ifile, service_parmeters)      

        # if necessary, localize file
        localized_file = localize_file(this_file['src'], servicefile_path)

        # plan to render the file into a 'tmp' directory under the servicefile dir
        rendered_file_path = os.path.join(servicefile_path,'tmp')

        if os.path.exists(localized_file):

            # replace any service parmeters variables in the file body and return the 
            # name+path of the "rendered" file
            rendered_file = apply_templates_in_file(localized_file, service_parmeters, rendered_file_path)
  
            # if destination is s3 bucket
            if (is_s3_destination(this_file['dest']) and rendered_file):

                # copy rendered file to s3 destination 
                cp_file_to_s3( rendered_file, this_file['dest'])

            # if destination is another local file (mostly used for testing)
            else:

                # make sure destination directory exists
                if not os.path.exists(os.path.dirname(this_file['dest'])):
                    os.makedirs(os.path.dirname(this_file['dest']))

                # copy rendered file to the destination
                shutil.copy(rendered_file,this_file['dest']) 

        else:

            raise FileError( "%s file deploy was not performed. Source file is missing"%localized_file )


# Render service parmeters into file body, then load files to destination
# Only s3 destinations are currently supported.
def _load_files_old(files, service_parmeters):

    # assume the file path is relative to the location
    # of the service descriptor file (just like Dockerfile!)
    servicefile_path = get_variable(service_parmeters,"servicefile-path")

    for this_ifile in files:

        # render intrinsics in the file dictionary
        this_file = apply_fxn(this_ifile, service_parmeters)      

        source_path = os.path.join(servicefile_path,this_file['src'])

        # render the file into a 'tmp' directory under the servicefile dir
        rendered_file_path = os.path.join(servicefile_path,'tmp')

        if os.path.exists(source_path):

            # replace any service parmeters variables in the file body and return the 
            # name+path of the "rendered" file
            rendered_file = apply_templates_in_file(source_path, service_parmeters, rendered_file_path)
  
            # if destination is s3 bucket
            if (is_s3_destination(this_file['dest']) and rendered_file):

                # copy rendered file to s3 destination 
                cp_file_to_s3( rendered_file, this_file['dest'])

            # if destination is another local file (mostly used for testing)
            else:

                # make sure destination directory exists
                if not os.path.exists(os.path.dirname(this_file['dest'])):
                    os.makedirs(os.path.dirname(this_file['dest']))

                # copy rendered file to the destination
                shutil.copy(rendered_file,this_file['dest']) 

        else:

            raise FileError( "%s file deploy was not performed. Source file is missing"%source_path )

def _load_dirs(directories, service_parmeters):

    # assume the directory path is relative to the location
    # of the service descriptor file (just like Dockerfile!)
    servicefile_path = get_variable(service_parmeters,"servicefile-path","")

    for this_idir in directories:

        # render intrinsics in the file dictionary
        this_dir = apply_fxn(this_idir, service_parmeters)      

        source_path = os.path.join(servicefile_path,this_dir['src'])

        # render files into a 'tmp' directory under the servicefile dir
        rendered_dir_path = os.path.join(servicefile_path,'tmp',this_dir['src'])

        if os.path.exists(source_path):

            # replace any service parmeters variables in the file body and return the 
            # "rendered" file contents as a string
            apply_templates_in_dir(source_path, service_parmeters, rendered_dir_path)

            # if destination is s3 bucket
            if is_s3_destination(this_dir['dest']):

                # sync directories to s3 destination
                sync_dir_to_s3( rendered_dir_path, this_dir['dest'])

            # if destination is another local dir (mostly used for testing)
            else:

                # clear destination dir if it exists
                if os.path.exists(this_dir['dest']):
                    shutil.rmtree(this_dir['dest'])

                # recursively copy tmp dir to the destination
                shutil.copytree(rendered_dir_path,this_dir['dest']) 
        else:

            raise FileError( "%s directory deploy was not performed. Source dir is missing"%source_path )


# returns true if file to be loaded is configured for an s3 destination
def is_s3_destination( destination ):

    s3_destination = False

    # S3 destinations are URL's with s3 as the scheme
    # Use this to detect an S3 destination

    # attempt to parse the destination as a URL
    url_parts = urlparse.urlparse(destination)

    if (url_parts and url_parts.scheme and url_parts.scheme == 's3'):

        s3_destination = True

    return s3_destination

# cp a file to an s3 bucket
# raises an Error if source file does not exists
# or if s3 cp fails
def cp_file_to_s3(source_file, destination_s3_url):

    # make sure source file exists
    if os.path.exists(source_file):            

        # form aws cp command for this file
        aws_cmd = "aws s3 cp %s %s"%( source_file, destination_s3_url)

        try:
            subprocess.check_output( aws_cmd , stderr=subprocess.STDOUT, shell=True )

        except subprocess.CalledProcessError as e:
            raise FileError("Error copying file to s3 destination: %s"%e)

    else:
        raise FileError("Source file %s does not exist"%source_file)

# sync a directory to an s3 bucket
# raises an Error if source dir does not exists
# or if s3 sync fails
def sync_dir_to_s3(source_dir, destination_s3_url):

    # make sure source file exists
    if os.path.exists(source_dir):            

        # form aws sync command for this directory
        # use --delete option to remove any files or directories already in s3 
        # destination that aren't in source_dir
        aws_cmd = "aws s3 sync %s %s %s"%( source_dir, destination_s3_url,"--delete")

        try:
            subprocess.check_output( aws_cmd , stderr=subprocess.STDOUT, shell=True )

        except subprocess.CalledProcessError as e:
            raise FileError("Error copying directory to s3 destination: %s"%e)

    else:
        raise FileError("Source directory %s does not exist"%source_dir)

# Get the ip addresses of ec2 instances in a stack identified by stack_name
# with a name containing name_search_string.
# If no name_search_string provided, returns IP addresses of all stack EC2 instances
def get_ec2_ips( stack_name , name_search_string=""):

    # filters for stack_id in cloudformation tag
    filters = [{'Name':"tag:aws:cloudformation:stack-name", 'Values' : [stack_name]}]

    # if specifying a Name, filter on that name
    if name_search_string:
        filters.append({'Name':"tag:Name", 'Values' : [name_search_string]})

    #pull response with instances matching filters
    client = boto3.client('ec2')
    instances = client.describe_instances(Filters=filters)

    # get the private ip address of instances that are running
    ips = jmespath.search("Reservations[?Instances[?State.Name=='running']].Instances[*].PrivateIpAddress", instances)

    # because the of the reservations in the outer layer, the search above will return 
    # the IP array in a out array, e.g.
    # [[ip1, ip2, etc.]]

    # we need to get rid of outer array
    return ips[0]

# get the RDS endpoints associated with a stack
def get_rds_endpoints( stack_name ):

    endpoints = []

    # get the resources associated with the stack
    cloudformation = boto3.client('cloudformation')
    resources = cloudformation.describe_stack_resources(StackName=stack_name)

    rds_id = ""

    if 'StackResources' in resources:

        for resource in resources['StackResources']:

            if resource['ResourceType'] == 'AWS::RDS::DBInstance':

                rds_id = resource['PhysicalResourceId']

    if rds_id:

        client = boto3.client('rds')
        instances = client.describe_db_instances(DBInstanceIdentifier=rds_id)

        endpoints = jmespath.search("DBInstances[*].Endpoint.{Address: Address, Port: Port}", instances)

    return endpoints

# get the name of the elastic cache endpoint associated with a stack
def get_cache_endpoint( stack_name ):

    endpoint = {}

    # get the resources associated with the stack
    cloudformation = boto3.client('cloudformation')
    resources = cloudformation.describe_stack_resources(StackName=stack_name)

    cache_id = ""

    if 'StackResources' in resources:

        for resource in resources['StackResources']:

            if resource['ResourceType'] == 'AWS::ElastiCache::ReplicationGroup':

                cache_id = resource['PhysicalResourceId']

    if cache_id:

        client = boto3.client('elasticache')
        response = client.describe_replication_groups(ReplicationGroupId=cache_id)

        if ('ReplicationGroups' in response and len(response['ReplicationGroups'])>0):

            this_rep_group = response['ReplicationGroups'][0]

            for node in this_rep_group['NodeGroups']:

                if 'PrimaryEndpoint' in node:

                    endpoint = node['PrimaryEndpoint']

    return endpoint    

# get the subnets associated with an auto-scaling groups
def get_asg_subnet_ids( asg_name ):

    #pull response with instances matching filters
    client = boto3.client('autoscaling')
    response = client.describe_auto_scaling_groups(AutoScalingGroupNames=[asg_name])

    if response['AutoScalingGroups']:
        asg = response['AutoScalingGroups'][0]
    else:
        asg = {}

    subnet_ids = []

    if asg:
        subnet_id_str = asg['VPCZoneIdentifier']
        subnet_ids = subnet_id_str.split(',')
        
    return subnet_ids  

# get the iam role associated with an auto-scaling groups
def get_stack_iam_role( params ):

    asg_name = get_resource_name(params,'asg')

    #pull response with instances matching filters
    client = boto3.client('autoscaling')
    response = client.describe_auto_scaling_groups(AutoScalingGroupNames=[asg_name])

    if response['AutoScalingGroups']:
        asg = response['AutoScalingGroups'][0]
        launchConfigName = asg['LaunchConfigurationName']
    else:
        launchConfigName = ""

    iam_role = ""

    if launchConfigName:
        response = client.describe_launch_configurations(LaunchConfigurationNames=[launchConfigName])

        if response['LaunchConfigurations']:
            iam_role = response['LaunchConfigurations'][0]['IamInstanceProfile']
        
    return iam_role      

# get the ssh key associated with an auto-scaling groups
def get_stack_ssh_keys( params ):

    asg_name = get_resource_name(params,'asg')

    #pull response with instances matching filters
    client = boto3.client('autoscaling')
    response = client.describe_auto_scaling_groups(AutoScalingGroupNames=[asg_name])

    if response['AutoScalingGroups']:
        asg = response['AutoScalingGroups'][0]
        launchConfigName = asg['LaunchConfigurationName']
    else:
        launchConfigName = ""

    ssh_key = ""

    if launchConfigName:
        response = client.describe_launch_configurations(LaunchConfigurationNames=[launchConfigName])

        if response['LaunchConfigurations']:
            ssh_key = response['LaunchConfigurations'][0]['KeyName']
        
    return ssh_key   

# get the ssl cert associated with an elb
def get_stack_ssl_cert( params ):

    ielb_name = get_resource_name(params,'i-elb')
    eelb_name = get_resource_name(params,'e-elb')

    #pull response with instances matching filters
    client = boto3.client('elb')
    response = client.describe_load_balancers(LoadBalancerNames=[ielb_name,eelb_name])

    if response['LoadBalancerDescriptions']:
        elb = response['LoadBalancerDescriptions'][0]
    else:
        elb = {}

    ssl_cert = ""

    if elb:
        ssl_certs = jmespath.search('[*].Listener.SSLCertificateId',elb['ListenerDescriptions'])
        if (ssl_certs and len(ssl_certs)==1):
            ssl_cert = ssl_cert[0]
        
    return ssl_cert     

# returns true if an existing stack has external (public) access
def stack_has_external_access( params ):

    external_access = False

    eelb_name = get_resource_name(params,'e-elb')

    #pull response with instances matching filters
    client = boto3.client('elb')
    response = client.describe_load_balancers(LoadBalancerNames=[eelb_name])

    if response['LoadBalancerDescriptions']:
        elb = response['LoadBalancerDescriptions'][0]
    else:
        elb = {}

    # if the e-elb was found, stack provides external access
    if elb:
        external_access = True

    return external_access

# get vpc associated with an existing stack
def get_stack_vpc( params ):

    stack_name = get_stack_name(params)

    vpc_id = ""

    if stack_name:

        # get the stack
        cloudformation = boto3.client('cloudformation')

        try:
            stack = cloudformation.describe_stacks(StackName=stack_name)

            # get the first stack's 'stack id'
            stack_id = stack['Stacks'][0]['StackId']

            # filters for stack_id in cloudformation tag
            filters = [{'Name':"tag:aws:cloudformation:stack-id", 'Values' : [stack_id]}]

            #pull response with instances matching filters
            client = boto3.client('ec2')

            reservations = client.describe_instances(Filters=filters)

            intances = jmespath.search('Reservations[*].Instances',reservations)

            if len(intances)>=1:

                # use the vpd id of the first instance
                # print intances[0][0]
                vpc_id = intances[0][0]['VpcId']

                vpcs = client.describe_vpcs(VpcIds=[vpc_id])

                if 'Vpcs' in vpcs and len(vpcs['Vpcs'])==1:
                    vpc =  vpcs['Vpcs'][0]
                    vpc_id = vpc["VpcId"]

        except ClientError as e:
            print "existing stack not found"

        return vpc_id   

# get the subnets associated with an internal load balancer
def get_elb_subnet_ids( load_balancer_name ):

    #pull response with instances matching filters
    client = boto3.client('elb')
    response = client.describe_load_balancers(LoadBalancerNames=[load_balancer_name])

    if response['LoadBalancerDescriptions']:
        elb = response['LoadBalancerDescriptions'][0]
    else:
        elb = {}

    subnet_ids = []

    if elb:
        subnet_ids = elb['Subnets']
        
    return subnet_ids

def get_stack_param_value( stack_name , stack_param_name ):
    
    client = boto3.client('cloudformation')

    response = client.describe_stacks(StackName=stack_name)
    
    if response['Stacks']:
        stack = response['Stacks'][0]
    else:
        stack = {}

    value = ""

    if (stack and 'Parameters' in stack):
        for param in stack['Parameters']:
            if param['ParameterKey'] == stack_param_name:
                value = param['ParameterValue'] 
        
    return value

# get the value from a tag
def get_stack_tag_value( service_params , stack_tag_name ):

    stack_name = get_stack_name(service_params)
    
    client = boto3.client('cloudformation')

    response = client.describe_stacks(StackName=stack_name)
    
    if response['Stacks']:
        stack = response['Stacks'][0]
    else:
        stack = {}

    value = ""

    if (stack and 'Tags' in stack):
        for param in stack['Tags']:
            if param['Key'] == stack_tag_name:
                value = param['Value'] 
        
    return value
   


#!/usr/bin/env python

import argparse, getpass, jmespath, json

from yac.lib.container.push import push_image
from yac.lib.service import get_service, get_service_parmeters, get_service_alias
from yac.lib.stack import get_stack_templates, get_stack_name, get_ec2_ips, deploy_stack_files
from yac.lib.naming import set_namer
from yac.lib.container.config import get_aliases, env_ec2_to_api, get_volumes_map
from yac.lib.container.config import get_port_bindings
from yac.lib.container.api import get_connection_str
from yac.lib.vpc import get_vpc_prefs

def main():
    
    parser = argparse.ArgumentParser(description='Push container images from a remote host to docker hub.')

    # required args           
    parser.add_argument('servicefile', help='location of the Servicefile (registry key or abspath)')
    parser.add_argument('name',        help='name of container to push') 
    parser.add_argument('env',         help='the environment where the container image resides', 
                                       choices=  ['dev','stage','prod'])
    parser.add_argument('-a',
                        '--alias',  help='service alias for the stack currently supporting this service (deault alias is per Servicefile)')
    parser.add_argument('-d','--dryrun',  help='dry run the container start by printing rendered template to stdout', 
                                          action='store_true')
    parser.add_argument('-p',
                        '--params',  help='path to a file containing additional, static, service parameters (e.g. vpc params, or service constants)',
                                     default='') 
    args = parser.parse_args()

     # determine service defintion, complete service name, and service alias based on the args
    service_descriptor, service_name, servicefile_path = get_service(args.servicefile) 

    # get vpc preferences in place
    vpc_prefs = get_vpc_prefs()

    # set the resource namer to use with this service
    set_namer(service_descriptor, servicefile_path, vpc_prefs)

    # get the alias to use with this service
    service_alias = get_service_alias(service_descriptor,args.alias)

    # get the service parameters for use with yac-ref's in service templates
    service_parmeters = get_service_parmeters(service_alias, args.env, args.params, 
                                              service_name, service_descriptor,
                                              servicefile_path, vpc_prefs)

    # get cloud formation template for the service requested and apply yac intrinsic 
    # functions (yac-ref, etc.) using  the service_parmeters
    stack_template = get_stack_templates(service_descriptor,
                                         vpc_prefs,  
                                         service_parmeters)

    app_taskdefs = stack_template['Resources']['TaskDefs']['Properties']

    image_tag = jmespath.search("ContainerDefinitions[?Name=='%s'].Image"%args.name,app_taskdefs)[0]

    # get stack name
    stack_name = get_stack_name(service_parmeters)

    # get the ip address of the target host
    target_host_ips = get_ec2_ips( stack_name )

    if args.dryrun:

        # sanity check
        print('About to push image %s on %s to docker hub.'%(image_tag, target_host_ips))

    else:

        # get username/pwd/email for docker hub account
        hub_uname=raw_input('Hub username: ')
        hub_pwd=getpass.getpass('Hub password: ')
        hub_email=raw_input('Hub email: ')
    
        # start the container on each host ...
        for target_host_ip in target_host_ips:

            # get connection string for the docker remote api on the target host
            docker_api_conn_str = get_connection_str( target_host_ip )

            # sanity check
            raw_input('About to push image %s on %s to docker hub. Hit <enter> to continue...'%(image_tag, target_host_ip))


            # push the image
            push_image( image_tag,
                        hub_uname,
                        hub_pwd,
                        hub_email,
                        docker_api_conn_str )



#!/usr/bin/env python

import argparse, os, inspect, jmespath, json
from colorama import Fore, Style
from yac.lib.service import get_service, get_service_parmeters, get_service_alias
from yac.lib.stack import get_stack_templates, get_stack_name, get_ec2_ips, deploy_stack_files
from yac.lib.naming import set_namer
from yac.lib.container.start import start
from yac.lib.container.config import get_aliases, env_ec2_to_api, get_volumes_map
from yac.lib.container.config import get_port_bindings
from yac.lib.container.api import get_connection_str
from yac.lib.vpc import get_vpc_prefs


def main():

    parser = argparse.ArgumentParser(description='Start a container per the container defintion in the provided Servicefile')

    # required args                                         
    parser.add_argument('servicefile', help='location of the Servicefile (registry key or abspath)')
    parser.add_argument('name',        help='name of container to start') 
    parser.add_argument('env',         help='the environment where the container should be started', 
                                       choices=  ['dev','stage','prod'])

    parser.add_argument('-s', '--source', 
                        help='image source for this container (hub=dockerhub, local=image on host filesystem)', 
                        choices=['hub','local'], 
                        default='local')

    parser.add_argument('-a',
                        '--alias',  help='service alias for the stack currently supporting this service (deault alias is per Servicefile)')                                       
    parser.add_argument('-c', '--cmd', 
                        help='run this cmd instead of the stock container CMD (see associated Dockerfile)')
    parser.add_argument('-d','--dryrun',  help='dry run the container start by printing rendered template to stdout', 
                                          action='store_true')
    parser.add_argument('-p',
                        '--params',  help='path to a file containing additional, static, service parameters (e.g. vpc params, of service constants)',
                                     default='') 

    args = parser.parse_args()

     # determine service defintion, complete service name, and service alias based on the args
    service_descriptor, service_name, servicefile_path = get_service(args.servicefile) 

    # abort if service_descriptor was not loaded successfully
    if not service_descriptor:
        print("The Servicefile input does not exist locally or in registry. Please try again.")
        exit()

    # get vpc preferences in place
    vpc_prefs = get_vpc_prefs()

    # set the resource namer to use with this service
    set_namer(service_descriptor, servicefile_path, vpc_prefs)

    # get the alias to use with this service
    service_alias = get_service_alias(service_descriptor,args.alias)

    # get the service parameters for use with yac-ref's in service templates
    service_parmeters = get_service_parmeters(service_alias, args.env, args.params, 
                                              service_name, service_descriptor,
                                              servicefile_path, vpc_prefs)

    # get cloud formation template for the service requested and apply yac intrinsic 
    # functions (yac-ref, etc.) using  the service_parmeters
    stack_template = get_stack_templates(service_descriptor,
                                         vpc_prefs,  
                                         service_parmeters)

    app_taskdefs = stack_template['Resources']['TaskDefs']['Properties']

    image_tags = jmespath.search("ContainerDefinitions[?Name=='%s'].Image"%args.name,app_taskdefs)
    if len(image_tags)!=1: 
        print("The container requested doesn't exist in the Servicefile. " +
              "Available containers include: %s."%jmespath.search("ContainerDefinitions[*].Name",app_taskdefs))
        exit()
    else:
        image_tag = image_tags[0]

    # get stack name
    stack_name = get_stack_name(service_parmeters)

    # get the ip address of the target host
    target_host_ips = get_ec2_ips( stack_name )

    if jmespath.search("ContainerDefinitions[?Name=='%s'].Environment"%args.name,app_taskdefs):
        ecs_env = jmespath.search("ContainerDefinitions[?Name=='%s'].Environment"%args.name,app_taskdefs)[0]
        env_variables = env_ec2_to_api(ecs_env)
    else:
        env_variables = {}

    # get the volumes map
    all_volumes = jmespath.search("Volumes",app_taskdefs)
    mount_points = jmespath.search("ContainerDefinitions[?Name=='%s'].MountPoints"%args.name,app_taskdefs)[0]
    volumes_map, volumes_bindings = get_volumes_map(all_volumes,mount_points)

    # get the volumes
    # volumes_bindings = jmespath.search("volumes[*].host.sourcePath",app_taskdefs)

    # get the port bindings for this container
    port_mappings = jmespath.search("ContainerDefinitions[?Name=='%s'].PortMappings"%args.name,app_taskdefs)[0]
    port_bindings = get_port_bindings(port_mappings)

    source = 'local' if args.source=='local' else 'remote (hub)'

    if args.dryrun:

        print(Fore.GREEN)

        user_msg = "Starting the %s container on %s using the %s image %s"%(args.name,
                                                                            target_host_ips,
                                                                            source, 
                                                                            image_tag)
        if args.cmd:
            user_msg = user_msg + "\nUsing command: %s"%args.cmd

        # do a dry-run by printing the stack template and stack parameters to stdout
        print "environment variables ... \n%s"%json.dumps(env_variables,indent=2)
        print "volumes map ... \n%s"%json.dumps(volumes_map,indent=2)
        print "volumes bindings ... \n%s"%json.dumps(volumes_bindings,indent=2)
        print "port bindings ... \n%s"%json.dumps(port_bindings,indent=2)
        print user_msg

        print(Style.RESET_ALL)

    else:

        # deploy any files specified by the service
        deploy_stack_files(service_descriptor, service_parmeters)

        # start the container on each host ...
        for target_host_ip in target_host_ips:

            print(Fore.GREEN)

            user_msg = "Starting the %s container on %s using the %s image %s"%(args.name,
                                                                                target_host_ip,
                                                                                source, 
                                                                                image_tag)

            if args.cmd:
                user_msg = user_msg + "\nUsing command: %s"%args.cmd

            raw_input(user_msg + "\nHit <enter> to continue..." )

            print(Style.RESET_ALL) 

            # get connection string for the docker remote api on the target host
            docker_api_conn_str = get_connection_str( target_host_ip )

            # start the container
            start(
                image_tag=image_tag,
                    envs=env_variables,
                    alias=args.name,
                    volume_mapping=volumes_map,
                    volumes_bindings=volumes_bindings,
                    port_bindings=port_bindings,
                    connection_str=docker_api_conn_str,
                    start_cmd=args.cmd,
                    image_source=args.source,
                    template_vars={},
                    files_to_load=[],
                    volumes_from=[],
                    create_only=False) 



from setuptools import setup

def readme():
    with open('README.rst') as f:
        return f.read()

setup(name='yac',
      version='0.92',
      description='yac portable services',
      long_description=readme(),
      classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Topic :: Text Processing :: Linguistic',
      ],
      keywords='service cloudformation aws ecs docker',
      url='https://bitbucket.org/thomas_b_jackson/yac',
      author='Thomas Jackson',
      author_email='thomas.b.jackson@gmail.com',
      license='MIT',
      packages=['yac'],
      scripts=['yac/bin/yac'],
      include_package_data=True,
      install_requires=[
          'boto3',
          'boto',
          'redis',
          'docker-py'      
      ],
      zip_safe=False)



import argparse,sys,os
from yac.lib.paths import get_root_path
from sets import Set

yac_commands = ['stack', 'service', 'prefs', 'primer', 'container']

def show_primer(command_array):

    path_elements = [get_root_path(),'primer']

    path_elements = path_elements + command_array

    # make sure no options are mixed in
    if len(Set(command_array) & Set(yac_commands)) != len(command_array):

        print ("Options cannot be used with the primer command. usage: yac <command> <subcommand> primer")

    else:
        with open(os.path.join( *path_elements)) as primer_file:
            primer_file_content = primer_file.read()

        print primer_file_content



import os, json, jmespath, copy
from sets import Set
from yac.lib.registry import get_registry_keys, get_remote_value, clear_entry_w_challenge
from yac.lib.registry import set_remote_string_w_challenge
from yac.lib.file import get_file_contents, register_file,get_file_abs_path
from yac.lib.file_converter import convert_local_files, find_and_delete_remotes
from yac.lib.variables import get_variable, set_variable
from yac.lib.intrinsic import apply_custom_fxn
from yac.lib.vpc import get_vpc_prefs
from yac.lib.validator import validate_dictionary
from yac.lib.stack import get_stack_name, stack_exists
from yac.lib.inputs import get_inputs, get_user_inputs

NULL_SERVICE = "service does not exist"

REQUIRED_FIELDS = ["service-name.value",
                 "default-alias.value",
                 "service-description.summary",
                 "service-description.details",
                 "service-description.maintainer.name",
                 "service-description.maintainer.email"]

SUPPORTED_KEYS = [ "service-name",
                 "service-inputs",
                 "default-alias",
                 "service-description",
                 "services-consumed",
                 "resource-namer",
                 "deploy-for-boot",
                 "service-params",
                 "post-function",
                 "stack-template"]

YAC_SERVICE_SUFFIX="-service"

class ServiceError():
    def __init__(self, msg):
        self.msg = msg

def validate_service(service_descriptor, servicefile_path):

    # validate the main services
    val_errors = validate_dictionary(service_descriptor, SUPPORTED_KEYS, REQUIRED_FIELDS)
    
    # get the services consumed, as specified in the server parameters
    if 'services-consumed' in service_descriptor and 'services' in service_descriptor['services-consumed']:

        # add each sub-service specified
        for service_key in service_descriptor['services-consumed']['services']:

            sub_service_name = get_variable(service_descriptor['services-consumed']['services'],service_key)        

            # retrieve the sub-service descriptor
            sub_service_descriptor,a,b = get_service(sub_service_name, servicefile_path)

            sub_service_errors = validate_dictionary(sub_service_descriptor, SUPPORTED_KEYS, REQUIRED_FIELDS)
            val_errors.update(sub_service_errors)
    
    return val_errors


def get_all_service_names():

    service_names = []

    # get all registry keys
    registry_keys = get_registry_keys()

    # find all keys with _naming suffix
    for key in registry_keys:

        if YAC_SERVICE_SUFFIX in key:
            # remove the suffix and append to list
            service_names = service_names + [key.replace(YAC_SERVICE_SUFFIX,'')]

    return service_names   

# determine service alias, service key, and service defintion based on the stack cli args
# second argument allows this fxn to be called recursively in support of nested services
def get_service(servicefile_arg, servicefile_path=""):

    # treat the arg as a path to a local file
    this_service_descriptor, service_name, servicefile_path = get_service_from_file(servicefile_arg, servicefile_path)

    if service_name == NULL_SERVICE:

        # Service does not exist as a local file

        # Treat servicefile_arg as the service name. See if it is complete (i.e.
        # includes a version label)
        if is_service_name_complete(servicefile_arg):

            # name is complete
            service_name = servicefile_arg

        # Treat servicefile_arg is a service name that lacks a version
        elif is_service_available_partial_name(servicefile_arg):

            # get complete name
            service_name = get_complete_name(servicefile_arg)

        # pull the service from the registry
        this_service_descriptor, service_name = get_service_by_name(service_name)

    # see if service has sub-services
    if 'services-consumed' in this_service_descriptor and 'services' in this_service_descriptor['services-consumed']:

        # add each sub-service specified
        for service_key in this_service_descriptor['services-consumed']['services']:

            sub_service_name = get_variable(this_service_descriptor['services-consumed']['services'],service_key)

            sub_service_comment = this_service_descriptor['services-consumed']['services'][service_key]['comment']

            print "adding '%s' services from %s ..."%(sub_service_comment,sub_service_name)

            # do a recursive call to retrieve the sub-service descriptor
            # sub's can't change service_name or servicefile_path, so return them to placeholders (a,b) ...
            sub_service_descriptor,a,b = get_service(sub_service_name, servicefile_path)

            update_service_descriptor(this_service_descriptor, sub_service_descriptor)

    return this_service_descriptor, service_name, servicefile_path

def get_service_from_file(servicefile_arg, servicefile_path=""):

    service_descriptor = {}
    service_name = NULL_SERVICE
    abs_path = ""

    file_contents = get_file_contents(servicefile_arg, servicefile_path)

    # pull the service descriptor from file
    if file_contents:

        service_descriptor = json.loads(file_contents)

        # pull the service name out of the descriptor
        service_name = get_variable(service_descriptor, 'service-name')

        # the servicefile_arg could be relative to servicefile_path (if 
        # this is a local sub-service) or absolute
        # determine the absolute path for either condition
        abs_path = get_file_abs_path(servicefile_arg, servicefile_path)       

    # initialize mergeable fields to defaut values
    init_mergeable_fields(service_descriptor)

    return service_descriptor, service_name, abs_path

def get_service_by_name(service_name):

    service_descriptor = {}

    if service_name:

        reg_key = service_name + YAC_SERVICE_SUFFIX

        # look in remote registry
        service_contents = get_remote_value(reg_key)

        if service_contents:

            # load into dictionary
            service_descriptor = json.loads(service_contents)

        else:
            # service doesn't exist
            service_name = NULL_SERVICE

    # initialize mergeable fields to defaut values
    init_mergeable_fields(service_descriptor)            

    return service_descriptor, service_name

def update_service_descriptor(service_descriptor, sub_service_descriptor):

    # merge inputs
    service_descriptor["service-inputs"]["inputs"].update(sub_service_descriptor["service-inputs"]["inputs"])
    service_descriptor["service-inputs"]["conditional-inputs"].update(sub_service_descriptor["service-inputs"]["conditional-inputs"])

    # merge params
    service_descriptor["service-params"].update(sub_service_descriptor["service-params"])

    # merge stack Parameters
    service_descriptor["stack-template"]["Parameters"].update(sub_service_descriptor["stack-template"]["Parameters"])

    # merge stack Resources
    service_descriptor["stack-template"]["Resources"].update(sub_service_descriptor["stack-template"]["Resources"])

    # merge stack Conditions
    service_descriptor["stack-template"]["Conditions"].update(sub_service_descriptor["stack-template"]["Conditions"])


# if "mergeable" fields are empty, initi them to null in order to make descriptor mergers easier
def init_mergeable_fields(service_descriptor):

    if "service-params" not in service_descriptor:
        service_descriptor["service-params"] = {}

    if "service-inputs" not in service_descriptor:
        service_descriptor["service-inputs"] = {
                                                "inputs": {},
                                                "conditional-inputs": {}
                                                }

    # stack Parameters
    if "stack-template" not in service_descriptor:
        service_descriptor["stack-template"] = {}

    if "Parameters" not in service_descriptor["stack-template"]:
        service_descriptor["stack-template"]["Parameters"] = {}

    if "Resources" not in service_descriptor["stack-template"]:

        service_descriptor["stack-template"]["Resources"] = {}

    if "Conditions" not in service_descriptor["stack-template"]:
        service_descriptor["stack-template"]["Conditions"] = {}


def clear_service(service_name, challenge):

    # if service is in fact registered
    service_descriptor = get_service_by_name(service_name)
    if service_descriptor:

        # clear service entry registry
        reg_key = service_name + YAC_SERVICE_SUFFIX
        clear_entry_w_challenge(reg_key, challenge)

        # clear any files referenced 
        find_and_delete_remotes(service_descriptor, challenge)      
    
    else:
        raise ServiceError("service with key %s doesn't exist"%service_name)

# register service into yac registry
def register_service(service_name, service_path, challenge):

    if os.path.exists(service_path):

        service_contents_str = get_file_contents(service_path)

        if service_contents_str:

            reg_key = service_name + YAC_SERVICE_SUFFIX

            # get the base path of the service file
            servicefile_path = os.path.dirname(service_path)

            updated_service_contents_str = convert_local_files(service_name,
                                                  service_contents_str,
                                                  servicefile_path,
                                                  challenge)

            # set the service in the registry
            set_remote_string_w_challenge(reg_key, updated_service_contents_str, challenge)

    else:
        raise ServiceError("service path %s doesn't exist"%service_path)

# service_name formatted as:
#  <organization>/<service>:<version>
# service_path is path to the file container the service descriptor
def publish_service_description(service_name, service_path):

    print "stub for service publication to human-readable docs - not yet implemented"

def is_service_alias(service_alias, vpc_prefs):

    is_alias = False

    # see if alias is a name in our vpc_prefs alias dict
    if "aliases" in vpc_prefs and service_alias in vpc_prefs['aliases']:
        is_alias = True

    return is_alias

def get_alias_from_name(complete_service_name):

    alias = ""

    if complete_service_name:

        name_parts = complete_service_name.split(':')

        # see if first part can be further split
        name_prefix_parts = name_parts[0].split('/')

        # treat the alias as the last of the prefix parts
        alias = name_prefix_parts[-1]

    return alias

def get_service_name(service_alias, vpc_prefs):

    server_name = ""
    # see if alias is a name in our vpc_prefs alias dict
    if "aliases" in vpc_prefs and service_alias in vpc_prefs['aliases']:
        server_name = vpc_prefs['aliases'][service_alias]

    return server_name 

# a service name is considered complete if it includes a version tag
def is_service_name_complete(service_name):

    is_complete = False

    name_parts = service_name.split(':')

    if len(name_parts)==2:

        # a tag is included, so name is complete
        is_complete = True

    return is_complete  

# if only know partial service name (no label), returns true
# if the complete version of the service is in registry
def is_service_available_partial_name(service_partial_name):

    is_available = False

    if not is_service_name_complete(service_partial_name):
        # see if a service with tag=latest is available in the registry
        complete_name_candidate = '%s:%s'%(service_partial_name,"latest")
        service_desc = get_service_by_name(complete_name_candidate)

        if service_desc:
            is_available = True

    return is_available

def get_complete_name(service_name):

    complete_name = ""

    if not is_service_name_complete(service_name):
        # see if a service with tag=latest is available in the registry
        complete_name_candidate = '%s:%s'%(service_name,"latest")
        service_desc = get_service_by_name(complete_name_candidate)

        if service_desc:
            complete_name = complete_name_candidate

    return complete_name

def get_service_alias(service_descriptor,service_alias_arg):

    # if service alias was provided in cli, use that as the alias.
    # o.w., use the default alias specified in the servicefile
    if service_alias_arg:
        service_alias = service_alias_arg
    else:
        service_alias = get_variable(service_descriptor,"default-alias")

    return service_alias

def get_service_parmeters(service_alias, params_file, 
                          service_name, service_descriptor,
                          servicefile_path, vpc_prefs={}):

    service_parmeters = {}

    # add params set via cli
    set_variable(service_parmeters,"service-alias",service_alias)
    set_variable(service_parmeters,"service-name",service_name)
    set_variable(service_parmeters,"servicefile-path",servicefile_path)

    # update with static vpc preferences 
    if "vpc-params" in vpc_prefs:
        service_parmeters.update(vpc_prefs["vpc-params"])

    # update with static service params
    if "service-params" in service_descriptor:
        service_parmeters.update(service_descriptor["service-params"])

    # update with static params specified via params file
    if params_file:
        params_from_file_str = get_file_contents(params_file)
        params_from_file = json.loads(params_from_file_str)
        service_parmeters.update(params_from_file)
        
    # add services consumed from the service descriptor to the params
    services_consumed = get_variable(service_descriptor,'services-consumed',[])
    set_variable(service_parmeters,'services-consumed',services_consumed)

    # add service description from the service summary to params
    stack_description = service_descriptor['service-description']['summary']
    set_variable(service_parmeters,'service-description',stack_description)
        
    # add service-inputs (used to drive user-prompts for params)
    if "service-inputs" in service_descriptor:        
        service_parmeters["service-inputs"] = service_descriptor["service-inputs"]

    # create dynamic vpc params via an (optional) set of service-inputs
    if "service-inputs" in vpc_prefs:
        
        # initialize a separate set of service params for use in vpc inputs
        # script. This will allow us to differentiate preferece-generated 
        # inputs from other inputs
        vpc_defaults = service_parmeters.copy()
        vpc_defaults['service-inputs'] = vpc_prefs["service-inputs"]

        vpc_inputs_script = get_variable( vpc_prefs, "service-inputs","")

        if vpc_inputs_script:
            # gather inputs via custom script
            apply_custom_fxn(vpc_inputs_script, vpc_defaults)
        else:
            # gather inputs
            get_inputs(vpc_defaults)

        # save the vpc defaults back to the params
        set_variable(service_parmeters,'vpc-defaults',vpc_defaults)

        # Inputs entered by user should be treated the same as input provided by
        # a params file, so save them into the stack params for downstream use.
        service_parmeters.update(get_variable(vpc_defaults,'user-inputs'))

    # create dynamic service params via an (optional) service inputs script
    if "service-inputs" in service_descriptor:

        service_input_script = get_variable(service_descriptor, "service-inputs","")

        if service_input_script:
            # gather inputs via custom script
            apply_custom_fxn(service_input_script, service_parmeters)
        else:
            # gather inputs
            get_inputs(service_parmeters)

    # merge user inputs collected from servicefile and render into string that can be used 
    # in a Stack parameters value
    # Strip out the comments fields to preserve space (this string will be saved to a string in stack parameters, and
    # and stack params are limitted in length)
    user_inputs = get_user_inputs(service_parmeters)

    inputs_copy = copy.deepcopy(user_inputs)

    set_variable(service_parmeters,
                'user-inputs-str', 
                json.dumps(prune_inputs(inputs_copy)), 
                "Inputs provided by users, as a string")

    # add default values for vpc-related params if none were provided
    default_vpc_params(service_parmeters)

    # add stack name and state 
    stack_name = get_stack_name(service_parmeters)
    set_variable(service_parmeters,'stack-name',stack_name, "Name of the service stack")
    set_variable(service_parmeters,'stack-exists',stack_exists(stack_name), "Stack exists")

    return service_parmeters

def prune_inputs(user_inputs):

    # remove comment field
    for key in user_inputs:
        if 'comment' in user_inputs[key]:
            user_inputs[key].pop('comment')

    return user_inputs

# set default values for vpc-related params (proxy, corporate cidr, etc.)
def default_vpc_params(service_parmeters):

    if not get_variable(service_parmeters,'proxy-port',""):
        set_variable(service_parmeters,'proxy-port',"")

    if not get_variable(service_parmeters,'proxy-cidr',""):
        set_variable(service_parmeters,'proxy-cidr',"")

    if not get_variable(service_parmeters,'corporate-cidr',""):
        set_variable(service_parmeters,'corporate-cidr',"")

    if not get_variable(service_parmeters,'dns-cidr',""):
        set_variable(service_parmeters,'dns-cidr',"0.0.0.0/0")        

    if not get_variable(service_parmeters,'ntp-servers',""):
        set_variable(service_parmeters,'ntp-servers',"0.pool.ntp.org;1.pool.ntp.org;2.pool.ntp.org;3.pool.ntp.org")     


import argparse,sys,os
import yac.cli.stack
import yac.cli.service
import yac.cli.db
import yac.cli.prefs
import yac.cli.container
import yac.cli.ssh

from yac.cli.primer import show_primer

def main():

    # first argument is help
    if (len(sys.argv)==1 or sys.argv[1] == '-h'):

        show_primer(['primer'])

    # last argument is primer
    elif sys.argv[len(sys.argv)-1] == 'primer':

        # show primer instructions
        show_primer(sys.argv[1:])

    else:

        # strip command from args list
        command = sys.argv[1]
        sys.argv = sys.argv[1:]

        if command == 'stack':

            return yac.cli.stack.main()

        elif command == 'service':

            return yac.cli.service.main()

        elif command == 'prefs':

            return yac.cli.prefs.main()    

        elif command == 'container':

            return yac.cli.container.main()    

        elif command == 'ssh':

            return yac.cli.ssh.main()
        else:

            return "command not supported, or not yet implemented"
        


import boto3, jmespath, sys, json, requests
from sets import Set
from yac.lib.variables import get_variable, set_variable
from yac.lib.naming import get_stack_name
from yac.lib.stack import stack_exists, get_asg_subnet_ids, get_stack_param_value
from yac.lib.stack import get_stack_vpc, get_stack_ssh_keys
             
USER_INPUTS_PARAM_KEY = "UserInputs"

def get_inputs(params, inputs_key='service-inputs'):

    user_inputs = {}

    # get input params
    if inputs_key in params and 'inputs' in params[inputs_key]:

        service_param = params[inputs_key]['inputs']

        for param_key in service_param:

            # handle this input
            value = handle_service_input(params, service_param[param_key], param_key)

            # accumulate this input
            set_variable(user_inputs,param_key,value, service_param[param_key]['help'])

    # get conditional input params
    if inputs_key in params and 'conditional-inputs' in params[inputs_key]:
        conditional_params = params[inputs_key]['conditional-inputs']
        
        for param_key in conditional_params:

            if 'input' in conditional_params[param_key]['condition']:

                # get the conditional check
                conditional_input = conditional_params[param_key]['condition']['input']

                conditional_value = ""
                if 'value' in conditional_params[param_key]['condition']:
                    conditional_value = conditional_params[param_key]['condition']['value']

                # get state of the input
                conditional_setting = get_variable(params,conditional_input,"")

                # print "input %s, condition setting: %s, condition value: %s"%(param_key,conditional_setting,conditional_value)

            elif 'fxn' in conditional_params[param_key]['condition']:

                # get the conditional check
                conditional_fxn = str(conditional_params[param_key]['condition']['fxn'])

                conditional_value = ""
                if 'value' in conditional_params[param_key]['condition']:
                    conditional_value = conditional_params[param_key]['condition']['value']

                # execute the conditional fxn
                conditional_setting = eval(conditional_fxn)(params)                

            # if no conditional value specified, or a value was specified and the 
            # input/fxn value matches the condition value, then
            # handle the service input (o.w. skip)
            if ( ('value' not in conditional_params[param_key]['condition']) or 
                 (conditional_value == conditional_setting) ):

                # print "handling %s ..."%param_key

                # handle this conditional input
                value = handle_service_input(params, conditional_params[param_key], param_key)

                # accumulate the resulting input
                set_variable(user_inputs,param_key,value, conditional_params[param_key]['help'])

    # convert inputs into a string and save to a variable that can be connected to the stack
    # for reference by 'yac stack' during stack updates
    set_variable(params,'user-inputs',user_inputs)


def handle_service_input(params, dynamic_param, param_key):

    wizard_arg = ""                
    if 'arg' in dynamic_param['wizard_fxn']:
        wizard_arg = dynamic_param['wizard_fxn']['arg']

    options = []
    if 'options' in dynamic_param['constraints']:
        options = dynamic_param['constraints']['options']

    elif 'options_fxn' in dynamic_param['constraints']:
        # call the spec'd options fxn to generate the list of options

        options_fxn = dynamic_param['constraints']['options_fxn']['name']

        if 'namespace' in dynamic_param['constraints']['options_fxn']:
            name_space = dynamic_param['constraints']['options_fxn']['name_space']
            # do an import in the options fxn
            import_statement = "from %s import %s"%(name_space,options_fxn)
            exec(import_statement)
               
        # next evaluate the options fxn
        if 'arg' in dynamic_param['constraints']['options_fxn']:
            options_arg = dynamic_param['constraints']['options_fxn']['arg']
            options = eval(options_fxn)(params,options_arg)
        else:
            options = eval(options_fxn)(params)

    required = True
    if 'required' in dynamic_param['constraints']:
        required = dynamic_param['constraints']['required']

    value = set_service_input(params, 
                     param_key, 
                     dynamic_param['description'], 
                     dynamic_param['help'],
                     dynamic_param['wizard_fxn']['name'],
                     required,
                     options,
                     wizard_arg)

    return value

# set an input in params
def set_service_input(params, 
                     param_key, 
                     param_desc,
                     param_help, 
                     wizard_fxn,
                     required,
                     options, 
                     wizard_arg=""):

    value = ""

    # see if the param is already defined statically
    value = get_variable(params, param_key,'absolutely-nada')

    if value == 'absolutely-nada':

        # get the name of the stack
        stack_name = get_stack_name(params)

        # see if a vpc default value exists
        default_value = get_vpc_default(params, param_key)

        if stack_exists(stack_name) and not default_value:

            print "key: %s, value: %s"%(param_key,get_variable(params, param_key,"empty"))

            # a stack for this service already exists ...

            # determine the param value currently in use
            value = get_stack_service_inputs_value(stack_name, param_key)
 
            # give the user the choice of changing existing value
            if change_wizard(value, param_desc):
                value = eval(wizard_fxn)(param_desc,param_help,options,required,wizard_arg)

        elif not stack_exists(stack_name) and not default_value:

            # this is a new stack and no default value for this param is available ...

            # prompt user to provide a value for this param
            value = eval(wizard_fxn)(param_desc,param_help,options,required,wizard_arg)

        elif not stack_exists(stack_name) and default_value:

            # use the default
            print "setting %s to %s per vpc prefs"%(param_key,default_value)
            value = default_value

        elif stack_exists(stack_name) and default_value:

            # use the default
            print "keeping %s as %s per vpc prefs"%(param_key,default_value)
            value = default_value            

        set_variable(params, param_key,value, param_desc)

    return value

# get static or vpc-default param value
def get_variable_or_default(params, param_key):

    # see if the param is already defined statically
    value = get_variable(params, param_key)

    if not value:

        # see if the param is defined as a vpc-default
        vpc_defaults = get_variable(params,'vpc-defaults', {})

        value = get_variable(vpc_defaults,param_key)

        if value:
            # set this variable per the vpc default
            print 'setting %s per vpc default value: %s'%(param_key,value)
            set_variable(params,param_key,value)

    return value

def does_stack_exist(params):

    stack_name = get_stack_name(params)

    return stack_exists(stack_name)

# get param value set via vpc prefs
def get_vpc_default(params, param_key):

    # see if the param is defined as a vpc-default via vpc prefs
    vpc_defaults = get_variable(params,'vpc-defaults', {})

    value = get_variable(vpc_defaults,param_key)

    return value

def get_user_inputs_as_stack_param(service_params):

    stack_parameter = {}

    # get inputs driven from service file
    user_inputs = get_user_inputs(service_params)

    if user_inputs:

        user_inputs_str = json.dumps(user_inputs)

        print user_inputs_str

        stack_parameter['ParameterKey'] = USER_INPUTS_PARAM_KEY
        stack_parameter['ParameterValue'] = user_inputs_str

    return stack_parameter

def get_user_inputs(service_params):

    stack_parameter = {}

    # get inputs driven from service file
    user_inputs = get_variable(service_params, 'user-inputs')

    # get inputs driven from vpc preferences
    if ('vpc-defaults' in service_params and 
        'user-inputs' in service_params['vpc-defaults']):
        user_inputs.update(service_params['vpc-defaults']['user-inputs'])
    
    return user_inputs

def string_wizard(param_desc, help_msg, options, required, str_len=""):

    value = validated_input(param_desc,
                             help_msg,
                             options,
                             string_validation,
                             required,
                             str_len)

    return value

def int_wizard(param_desc, help_msg, options, required, arg=""):

    value = validated_input(param_desc,
                            help_msg, 
                            options,
                            int_validation,
                            required)

    return value

def bool_wizard(param_desc, help_msg, options, required, arg=""):

    value = validated_input(param_desc,
                            help_msg, 
                            ["True", "False"],
                            bool_validation,
                            required)

    return value == "True"

def array_wizard(param_desc, help_msg, options, required, layer=""):
      
    value = validated_array_input(param_desc,
                                       help_msg,
                                       options,
                                       array_validation,
                                       required)

    return value


def change_wizard(current_value, field_name):

    msg = "Current '%s' value is %s. Do you want to change it? (y/n/<enter>) >> "%(field_name,current_value)

    validation_failed = True
    change = False

    # accept y, n, or empty string
    options = ['y','n', '']

    while validation_failed:

        input = raw_input(msg)

        # validate the input
        validation_failed = string_validation(input, options, False)

    # change iff 'y' was input
    if (input or input == 'y'):
        change = True

    return change

def validated_input(param_desc, help_msg, options, function, required=True, optional_arg=""):

    validation_failed = True

    if required:
        param_msg = "\nThis service requires the following input: %s"%param_desc
    else:
        param_msg = "\nThis service accepts the following optional input: %s"%param_desc

    print param_msg
    print help_msg

    if options:

        if len(options)>10:
            input = raw_input("Press <enter> to see a list of availble options >> ")
            print "Choices include: \n%s"%pp_list(options)
        else:
            print "Choices include: \n%s"%pp_short_list(options)

    while validation_failed:

        if options:
            input_msg = "Please paste in one of the available options for %s >> "%param_desc
        else:
            input_msg = "Please input a value for %s >> "%param_desc

        input = raw_input(input_msg)

        input = input.strip("'")

        # validate the input
        validation_failed = function(input, options, required, optional_arg)

    return input

def validated_array_input(param_desc, help_msg, options, function, required=True, optional_arg=""):

    validation_failed = True
    array_building = True

    if required:
        param_msg = "\nThis service requires the following input: %s"%param_desc
    else:
        param_msg = "\nThis service accepts the following optional input: %s"%param_desc

    print param_msg
    print pp_help(help_msg)

    if options:

        input = raw_input("Press <enter> to see a list of availble options >> ")
        print "Choices include: \n%s"%pp_list(options)

    print "Paste in values one at a time and press Enter ( press Enter when done ) ..."

    inputs = []

    while validation_failed:

        input = raw_input("... input an item for '%s' >> "%param_desc)

        if input:
            inputs.append(input)
        else:
            validation_failed,inputs = function(inputs, options, required, optional_arg)

    return inputs

def validated_array_input_old(param_desc, retry_msg, options, function):

    validation_failed = True
    array_building = True

    print "enter values one at a time, cr when done ..."

    inputs = []

    while validation_failed:

        input = raw_input(param_desc).strip("'")

        if input:
            inputs.append(input)
        else:
            validation_failed,inputs = function(inputs, options,retry_msg)

def string_validation(input,options,required,max_strlen=4000):

    validation_failed = False
    if options:
        
        if type(options[0])==dict:
            # pull out just the options values
            options = jmespath.search("[*].Option",options)

        if input not in options:
            validation_failed = True
            retry_msg = "Input invalid - please select from the available options"

    if max_strlen:
        if len(input)>max_strlen:
            validation_failed = True
            retry_msg = "Input invalid (too long) - please input a string with <= %s chars"%max_strlen

    if required and not input:
        validation_failed = True
        retry_msg = "Input required - please enter a value"

    if validation_failed:
        print retry_msg

    return validation_failed

def int_validation(input,options,required,max_value=sys.maxint):

    validation_failed = False
    if options:
        if input not in options:
            validation_failed = True
            retry_msg = "Input invalid - please select from the available options"

    if max_value:
        if input>max_value:
            validation_failed = True
            retry_msg = "Input invalid - please input an int <= %s"%max_value

    if required and not input:
        validation_failed = True
        retry_msg = "Input required - please enter a value"

    if ( input and not input.isdigit() ):
        validation_failed = True
        retry_msg = "Input invalid - integers only"
        
    if validation_failed:
        print retry_msg

    return validation_failed    

def bool_validation(input,options,required,optional_arg=""):

    validation_failed = (input not in ["True", "False"])

    if required and not input:
        retry_msg = "Input required - please enter a value"

    elif validation_failed:
        retry_msg = "Input invalid - True/False only"

    if validation_failed:
        print retry_msg

    return validation_failed  

def input_validation(input,options,required,retry_msg):

    # attempt to find the vpc input
    validation_failed = input not in options

    if validation_failed:
        print retry_msg

    return validation_failed

def array_validation(inputs,options,required,arg=""):

    validation_failed = len(Set(inputs) & Set(options)) != len(inputs)

    if validation_failed:
        retry_msg = "Input invalid - please select from the available options"
        print retry_msg
        inputs=[]

    return validation_failed, inputs    

# get the ids of the vpcs available to this user
def get_vpc_ids(params):

    ec2 = boto3.client('ec2')

    # Environment tag contains VPC_MAPPING values
    vpcs = ec2.describe_vpcs()

    # get id's only
    vpc_ids = jmespath.search("Vpcs[*].VpcId", vpcs)

    return vpc_ids

# get names of available sns topics buckets
def get_s3_buckets(params):

    s3 = boto3.client('s3')

    buckets = s3.list_buckets()

    # get names only
    bucket_names = jmespath.search("Buckets[*].Name", buckets)

    return bucket_names

# get names of available sns topic arns
def get_sns_topic_arns(params):

    s3 = boto3.client('sns')

    topics = s3.list_topics()

    topics_arns = jmespath.search("Topics[*].TopicArn", topics)

    return topics_arns

# get names of available key pairs
def get_key_pairs(params):

    s3 = boto3.client('iam')

    public_keys = s3.list_ssh_public_keys()

    # get names only
    key_names = jmespath.search("SSHPublicKeys[*].SSHPublicKeyId", public_keys)

    return key_names

# get names of available key pairs
def get_ssl_certs(params):

    iam = boto3.client('iam')

    certs = iam.list_server_certificates()

    # get names only
    cert_name = jmespath.search("ServerCertificateMetadataList[*].ServerCertificateName", certs)

    return cert_name

# get names of available iam roles
def get_iam_roles(params):

    iam = boto3.client('iam')

    profiles = iam.list_instance_profiles()

    # get id only
    profile_ids = jmespath.search("InstanceProfiles[*].InstanceProfileId", profiles)

    return profile_ids  

# get names of available db snapshots
def get_snapshot_ids(params,name_search_string):

    iam = boto3.client('rds')

    filters = [{'Name':"tag:Name", 'Values' : [name_search_string]}]

    snapshots = iam.describe_db_snapshots(Filters=filters)

    # get id only
    snapshots_ids = jmespath.search("DBSnapshots[*].DBSnapshotIdentifier", snapshots)

    return snapshots_ids        

# get the zones available to this user
def get_azs(params):

    ec2 = boto3.client('ec2')

    # Environment tag contains VPC_MAPPING values
    azs = ec2.describe_availability_zones(Filters=[{'Name': 'state','Values': ['available']}])

    # get id's only
    az_names = jmespath.search("AvailabilityZones[*].ZoneName", azs)

    return az_names  

# get CoreOS AMI suitable for use with Jira
def get_coreos_ami(params):

    ec2 = boto3.client('ec2')

    # only return CoreOS hvm releases on the stable channel
    images = ec2.describe_images(Filters=[{"Name": "name", "Values": ["CoreOS-stable*hvm"]},
                                          {"Name": "owner-id", "Values": ["595879546273"]}])

    # get image id and image name
    image_ids = jmespath.search("Images[*].{Option: ImageId, Description: Name}", images)

    return image_ids

# get the subnets available in a given vpc and a given set of 
# availability zones
def get_subnet_ids(params):

    vpc_id = get_variable(params,"vpc-id") 
    availabilty_zones = get_variable(params,"availability-zones") 

    ec2 = boto3.client('ec2')
    
    subnets = ec2.describe_subnets(Filters=[
        {
            'Name': 'vpc-id',
            'Values': [
                vpc_id,
            ]
        },
        {
            'Name': 'availability-zone',
            'Values': availabilty_zones
        }])

    subnet_ids = jmespath.search('Subnets[*].SubnetId',subnets)

    return subnet_ids

# get the id's of db subnet groups available in a given vpc and a given set of 
# availability zones
def get_db_subnet_group_ids(params):

    vpc_id = get_variable(params,"vpc-id") 

    rds = boto3.client('rds')
    
    subnets = rds.describe_db_subnet_groups()

    db_subnet_group_ids = jmespath.search("DBSubnetGroups[?VpcId=='%s'].DBSubnetGroupName"%vpc_id,subnets)

    return db_subnet_group_ids       

# find the versions (tags) available for a container image
def get_image_versions(params, image_name, registry_url='https://registry.hub.docker.com'):

    # endpoint use to determine the latest version in the registry input for this app
    endpoint_uri = "/v2/repositories/%s/tags"%(image_name)

    # hit the endpoint
    endpoint_response = requests.get(registry_url + endpoint_uri) 

    # use jmespath to extract just the version value for each
    versions = jmespath.search("results[*].name",endpoint_response.json())

    return versions

# get all service-inputs collected from the user at service installation time.
def get_stack_service_inputs( stack_name ):

    decoded_dict = {}

    stack_param_key = USER_INPUTS_PARAM_KEY

    param_value = get_stack_param_value (stack_name , stack_param_key) 

    if param_value:

        # decode the params dict
        decoded_dict = json.loads(param_value)

    return decoded_dict 

# get the value from a stack param that contains service-inputs collected
# from the user at service installation time.
def get_stack_service_inputs_value( stack_name , yac_param_key ):

    yac_param_value = ""

    stack_param_key = USER_INPUTS_PARAM_KEY

    param_value = get_stack_param_value (stack_name , stack_param_key) 

    if param_value:

        # decode the params dict
        decoded_dict = json.loads(param_value)
        if yac_param_key in decoded_dict:
            yac_param_value = get_variable(decoded_dict,yac_param_key)

    return yac_param_value 

def pp_list(list):
    str = ""
    for item in list:
        str = str + '* %s\n'%item

    return str

def pp_short_list(list):
    str = ""
    for item in list:
        str = str + ' %s'%item

    return "( %s )"%str    

def pp_help(help):

    if type(help)==list:
        ret = pp_list(help)
    else:
        ret = str(help)

    return ret



import os, urlparse
from yac.lib.registry import set_remote_string_w_challenge, get_remote_value, get_registry_keys
from yac.lib.registry import clear_entry_w_challenge
from yac.lib.registry import set_local_value, get_local_value, delete_local_value, get_local_keys
from yac.lib.paths import get_yac_path,get_lib_path


YAC_FILE_PREFIX="yac://"

class FileError():
    def __init__(self, msg):
        self.msg = msg

def get_all_file_keys():

    file_keys = []

    # get all registry keys
    registry_keys = get_registry_keys()

    # find all keys with _naming suffix
    for key in registry_keys:

        if file_in_registry(key):
            # remove the naming part
            file_keys = file_keys + [key.replace(YAC_FILE_PREFIX,'')]

    return file_keys    

def get_file_from_registry(file_key):

    file_contents = ""

    reg_key = get_file_reg_key(file_key)

    # get file from registry
    file_contents = get_remote_value(reg_key)

    return file_contents

def clear_file_from_registry(file_path, challenge):

    # if file is in fact registered
    if get_file_from_registry(file_path):

        # clear file entry
        reg_key = get_file_reg_key(file_path)

        clear_entry_w_challenge(reg_key, challenge)      
    
    else:
        raise FileError("file with key %s doesn't exist"%file_path)

# register file into yac registry
def register_file(file_key, file_path, challenge):

    if os.path.exists(file_path):

        with open(file_path) as file_path_fp:

            file_contents = file_path_fp.read()

            reg_key = get_file_reg_key(file_key)

            # set the file in the registry
            set_remote_string_w_challenge(reg_key, file_contents, challenge)

    else:
        raise FileError("file at %s doesn't exist"%file_path)

def get_file_reg_key(file_with_path, file_namespace=""):

    file_key = file_with_path
    if file_namespace:
        file_key = os.path.join(file_namespace,file_with_path)

    # add prefix to make it easy to identify files in the registry 
    return YAC_FILE_PREFIX + file_key

def get_file_contents(file_key_or_path, servicefile_path=""):
    
    file_contents = ""

    # if file is in registry
    if file_in_registry(file_key_or_path):
        file_contents = get_remote_value(file_key_or_path)

    # if file exists locally
    elif os.path.exists(file_key_or_path):
        with open(file_key_or_path) as file_arg_fp:
            file_contents = file_arg_fp.read()

    # if file exists relative to the service descriptor
    elif os.path.exists(os.path.join(servicefile_path,file_key_or_path)):
        with open(os.path.join(servicefile_path,file_key_or_path)) as file_arg_fp:
            file_contents = file_arg_fp.read()

    return file_contents

def get_file_abs_path(file_key_or_path, servicefile_path):
    
    abs_path = ""

    # if file exists locally
    if os.path.exists(file_key_or_path):
        abs_path = os.path.dirname(file_key_or_path)

    # if file exists relative to the service descriptor
    elif os.path.exists(os.path.join(servicefile_path,file_key_or_path)):
        abs_path = os.path.dirname(os.path.join(servicefile_path,file_key_or_path))

    return abs_path    

def localize_file(file_key_or_path, servicefile_path=""):

    # if file is in registry
    if file_in_registry(file_key_or_path):
        # convert the file to a local version
        file_contents = get_remote_value(file_key_or_path)
        localized_file = create_customization_file(file_key_or_path, file_contents)

    # if file exists locally
    elif os.path.exists(file_key_or_path):
        localized_file = file_key_or_path

    # if file exists relative to the service descriptor
    elif os.path.exists(os.path.join(servicefile_path,file_key_or_path)):
       localized_file = os.path.join(servicefile_path,file_key_or_path)

    return localized_file    

# a file is in the registry if the key includes the yac file prefix
def file_in_registry(file_key):
    return YAC_FILE_PREFIX in file_key

# a file is in yac sources if it exists in yac sources
def file_in_yac_sources(file_key):

    sources_root = get_yac_path()

    source_path = os.path.join(sources_root,file_key) 

    return os.path.exists(source_path)       

def create_customization_file(file_yac_url, file_contents=""):

    if not file_contents:
        file_contents = get_file_contents(file_yac_url)

    # yac file url's are like: yac://<netloc>/<path>
    file_parts = urlparse.urlparse(file_yac_url)

    # Write the script contents to a local file under yac/lib/customizations
    # Include the netloc path to prevent collisions with other custom scripts
    script_file_rel_path = file_parts.netloc + file_parts.path

    script_file_path = os.path.join( get_lib_path(),
                                     'customizations', 
                                     script_file_rel_path)

    # make the directory if it does not alread exist
    script_file_dir = os.path.dirname(script_file_path)
    if not os.path.exists(script_file_dir):
        os.makedirs(script_file_dir)

    # write the file
    with open(script_file_path,'w') as script_file_path_fp:
       script_file_path_fp.write(file_contents)

    return script_file_path    



import os, sys, json, argparse, getpass, inspect, pprint
from colorama import Fore, Style
from botocore.exceptions import ClientError
from yac.lib.file import FileError
from yac.lib.stack import create_stack, update_stack, cost_stack, get_stack_state
from yac.lib.stack import BUILDING, UPDATING, stack_exists, get_stack_templates
from yac.lib.stack import deploy_stack_files
from yac.lib.stack import UNKNOWN_STATE, UPDATABLE_STATES, NON_EXISTANT
from yac.lib.inputs import get_user_inputs, get_user_inputs_as_stack_param
from yac.lib.service import get_service, get_service_parmeters, get_service_alias, NULL_SERVICE
from yac.lib.naming import set_namer, get_stack_name
from yac.lib.variables import get_variable, set_variable
from yac.lib.intrinsic import apply_custom_fxn, INSTRINSIC_ERROR_KEY
from yac.lib.vpc import get_vpc_prefs

def main():

    parser = argparse.ArgumentParser('Print a yac service to your cloud')
    
    # required args     
    parser.add_argument('servicefile',   help='location of the Servicefile (registry key or abspath)')

    # optional
    # store_true allows user to not pass a value (default to true, false, etc.)    
    parser.add_argument('-p',
                        '--params',  help='path to a file containing static service parameters (useful for keeping stack in config mgmt)',
                                     default='') 
    parser.add_argument('-a',
                        '--alias',  help='override default service alias with this value (used for stack resource naming')    

    parser.add_argument('-d',
                        '--dryrun',  help='dry run the stack change, printing template to stdout', 
                                     action='store_true')                                                                          

    args = parser.parse_args()

    # determine service defintion, complete service name, and service alias based on the args
    service_descriptor, service_name, servicefile_path = get_service(args.servicefile) 

    # abort if service descriptor was not loaded successfully
    if service_name == NULL_SERVICE:
        print("The Servicefile input does not exist locally or in registry. Please try again.")
        exit()

    # get any vpc preferences in place
    vpc_prefs = get_vpc_prefs()

    # set the resource namer to use with this service
    set_namer(service_descriptor, servicefile_path, vpc_prefs)

    # get the alias to use with this service
    service_alias = get_service_alias(service_descriptor,args.alias)

    # get the service parameters based on all inputs
    service_parmeters = get_service_parmeters(service_alias, args.params, 
                                              service_name, service_descriptor,
                                              servicefile_path, vpc_prefs)

    # get stack name
    stack_name = get_stack_name(service_parmeters)  

    # get cloud formation template for the service requested and apply yac intrinsic 
    # functions (yac-ref, etc.) using  the service_parmeters
    stack_template = get_stack_templates(service_descriptor,
                                         service_parmeters)

    # Get any reference errors recorded in the service parameters. Each represents
    # a value that should have been rendered into the stack template, but wasn't.
    reference_errors = get_variable(service_parmeters,INSTRINSIC_ERROR_KEY,"")

    # get services consumed by the service
    services_consumed = get_variable(service_parmeters,'services-consumed',[])


    # **************************** Print Time! ************************************
 
     # determine if we are building or updating this stack
    action = UPDATING if stack_exists(stack_name) else BUILDING

    if args.dryrun:

        # show stack state
        stack_state = get_stack_state(stack_name)

        # pprint stack template to a string
        stack_template_str = json.dumps(stack_template,indent=2)

        print(Fore.GREEN)

        print "%s (dry-run) the %s service aliased as '%s'"%(action,service_name,service_alias)
        print "Stack state is currently: %s."%stack_state
        print "Service stack will be named: %s"%(stack_name)

        if reference_errors:
            pp = pprint.PrettyPrinter(indent=4)
            print "Service templates include reference errors. Each should be fixed before doing an actual print."
            print pp.pprint(reference_errors)

        print_template = raw_input("Print stack template to stdout? (y/n)> ")

        if print_template and print_template=='y':
            print "Stack template:\n%s"%stack_template_str
            print "Sanity check the template above."

        print_user_inputs = raw_input("Print user inputs to stdout? (y/n)> ")

        if print_user_inputs and print_user_inputs=='y':
            # get the user inputs as a stack parameter. 
            print "debug: getting inputs from cli ..."
            user_inputs = get_user_inputs(service_parmeters)
            print "User inputs:\n%s"%json.dumps(user_inputs,indent=2)
            print "Sanity check the inputs above."

        estimate_cost = raw_input("Estimate the cost of the resources required by this service? (y/n)> ")

        if estimate_cost and estimate_cost=='y':

            # calc the cost of this stack, and provide url showing calculation
            cost_response = cost_stack(template_string = stack_template_str, 
                                       stack_params    = [])
            print "Cost of the resources for this service can be viewed here: %s"%(cost_response)

        print(Style.RESET_ALL)                      

    else:

        # make sure stack is in an buildable or updatable state before proceeding
        stack_state = get_stack_state(stack_name)

        stack_actionable = (stack_state != UNKNOWN_STATE and 
                           ((stack_state in UPDATABLE_STATES) or (stack_state == NON_EXISTANT)))

        # for reals ...
        print(Fore.GREEN)

        if not reference_errors and stack_actionable:

            # dump stack template to a string
            stack_template_str = json.dumps(stack_template)
            
            print "%s the %s service aliased as '%s'"%(action,service_name,service_alias)
            print "Service stack will be named: %s"%(stack_name)

            # give user chance to bail
            raw_input("Hit <enter> to continue..." )

            print(Style.RESET_ALL) 

            # create the service stack
            try:

                # deploy any files specified by the service
                deploy_stack_files(service_descriptor, service_parmeters, servicefile_path)

                if action == BUILDING:
                    response  = create_stack(stack_name = stack_name,
                                             template_string = stack_template_str)

                else:
                     response = update_stack(stack_name = stack_name,
                                             template_string = stack_template_str)

                # if a post function is specified, call it now
                if get_variable(service_descriptor,'post-function'):
                    apply_custom_fxn(get_variable(service_descriptor,'post-function'), service_parmeters)
            
            except ClientError as e:
                print 'Service creation failed: %s'%str(e)

            except FileError as e:
                print 'Service creation failed: %s'%str(e.msg)                

        elif reference_errors:
            pp = pprint.PrettyPrinter(indent=4)
            print "Service templates include reference errors. Each must be fixed before a stack print can be performed."
            print pp.pprint(reference_errors)

        elif not stack_actionable:
            print "Stack is not currently actionable. Stack state is currently: %s."%stack_state

        print(Style.RESET_ALL)            



#!/usr/bin/env python

import os, argparse, getpass
#from slib.db import setup_db
#from slib.name import get_db_name, get_db_user_name, get_stack_name
#from slib.stack import get_rds_db_endpoint


def main():
	
	parser = argparse.ArgumentParser('Setup the DB for a new stack')
	# required args
	parser.add_argument('app',  help='the app name')
	parser.add_argument('env',  help='the env', choices=['dev', 'stage', 'prod', 'archive'])
	parser.add_argument('-x','--suffix',  help='app suffix, to support multiple instances of the same app in the same env')

	# pull out args
	args = parser.parse_args()


#!/usr/bin/env python

import argparse
from docker import Client
from yac.lib.container.api import get_docker_client, get_connection_str, find_container_by_name
from yac.lib.hosts import get_host

def main():
	
	parser = argparse.ArgumentParser(description='Stop one of the containers')

	# required args
	parser.add_argument('env',   help='the environment to be targeted', choices=['dev', 'stage', 'prod'])            
	parser.add_argument('type',  help='the host type to be targeted', choices=['rw', 'ro']) 
	parser.add_argument('alias', help='name (alias) of container to load', choices=container_aliases)

	# pull out args
	args = parser.parse_args()
	env = args.env
	host_type = args.type
	which_container = args.alias

	# get the ip of the targeted host
	target_host_ip = get_host(env, host_type, 'app')

	# get the ip address of the target host
	# the is currently only one sinstance per tack
	target_host_ip = get_ec2_ip( stack_name )[0]

	# get connection string for the docker remote api on the target host
	docker_api_conn_str = get_connection_str( target_host_ip )

	docker_client = get_docker_client( docker_api_conn_str )

	# find container
	container = find_container_by_name( args.alias, docker_api_conn_str )

	# stop the container
	# Provide a long timeout (30s) to give container time to gracefully shut itself down
	# After timeout expires, docker will sent SIGKILL which could interrupt an app's
	# shutdown sequence and could result in data integrity issues (e.g. the confluence
	# container writes data to its DB during shutdown. Interruping this process can result
	# in plugin data integrity issues, and can even prevent access to the UPM UI.)
	print docker_client.stop(container=container.get('Id'),
	                         timeout=60)



import requests, jmespath

# for the registry find, find the latest version of the app input, in the container org input
def get_latest_version(app, container_org, registry_url='https://registry.hub.docker.com'):

    # endpoint use to determine the latest version in the registry input for this app
    endpoint_uri = "/v2/repositories/%s/%s/tags"%(container_org,app)

    # hit the endpoint
    endpoint_response = requests.get(registry_url + endpoint_uri) 

    # use jmespath to extract just the version value for each
    versions = jmespath.search("results[*].name",endpoint_response.json())

    # jmespath leaves a weird prefix on each result. remove by str conversion
    latest_version = "not found!"

    if versions:

        versions = [str(i) for i in versions]

        latest_version = ""

        if (versions and len(versions)>=1):

            if 'latest' in versions:

                latest_version = "latest"
            
            else:
                
                # sort and grab the last value in the resulting list (the latest version)
                versions.sort()
                latest_version = versions[-1]

    return str(latest_version)

def get_app_version(app_alias, my_service_descriptor):

    app_version = ""

    # the version would be in the image for this app
    search_str = "task-definition.containerDefinitions[?name=='%s'].image"%app_alias
    images_found = jmespath.search(search_str, my_service_descriptor)

    if images_found and len(images_found)==1:

        # grab the version from the tail end of the image label
        app_version = str(images_found[0]).split(':')[-1]

    return app_version    


#!/usr/bin/env python

import os
from sqlalchemy import *
from sqlalchemy.orm import sessionmaker

# describes how to connect to a DB
# used for inputs to most functions below
class DB():

    def __init__(self, 
                 db_name,
                 db_user,
                 db_pwd,
                 db_host,
                 db_port,
                 db_engine):

        self.configs = {}

        self.configs['db_name'] = db_name
        self.configs['db_user'] = db_user
        self.configs['db_pwd']  = db_pwd
        self.configs['db_host'] = db_host
        self.configs['db_port'] = db_port        
        self.configs['engine'] = db_engine

# get the driver corresponding to the os we are running on
def get_db_driver(this_os = os.name):

    # drivers that seem to work best include:
    # windows:  pyodbc
    # macos:    pymssql

    if this_os == 'posix':
        # for posix-compliant os like macos
        driver = 'pymssql'
    else:
        # for windows this one seems to work best
        driver = 'pyodbc'

    return driver

# get connection string for connecting to an ms-sql RDBMS 
# db is an instance of the DB class
def get_connection_string(db):
    connection = ""
    driver = get_db_driver()
    if db.configs['engine']=='mssql':
        connection = "mssql+%s"%driver
    elif db.configs['engine']=='postgres':
        connection = 'postgresql'

    return "%s://%s:%s@%s/%s" % (connection, 
                                        db.configs['db_user'], 
                                        db.configs['db_pwd'], 
                                        db.configs['db_host'], 
                                        db.configs['db_name'])

# connect to database
# db is an instance of the DB class
def connect_to_db(db):
    return create_engine(get_connection_string(db))
    
# get an executable table (a table that alchemy commands can be run against)
# db is an instance of the DB class
def get_table(table_name, db):

    engine = connect_to_db(db)
    connection = engine.connect()
    metadata = MetaData(engine)

    # build table
    return Table(table_name, metadata, autoload=True)

def get_session(db):

    engine = connect_to_db(db)

    # create a configured "Session" class
    Session = sessionmaker(bind=engine)

    # create a Session
    session = Session()    


#!/usr/bin/env python

import sys
from docker import Client
from jinja2 import Template
import docker.utils, os, json, subprocess, urlparse
from api import get_docker_client, get_connection_str, find_container_by_name

# render template variable in the container environment variables.
# template_vars contains key/value pairs of template variables
# that may appear in container_envs values
def _render_env_variables(container_envs, template_vars):
    
    for key in container_envs:
        container_envs[key] = _render_string_variables(container_envs[key],
                                                       template_vars)
    return container_envs

def _render_string_variables(string_w_variables, template_vars):

    template = Template(string_w_variables)
    return template.render(template_vars)

# cp a file to an s3 buckert
# raises an Error if source file does not exists
# raises an subprocess.CalledProcessError if cp fails
def cp_to_s3(source_file, destination_s3_url):

    # make sure source file exists
    if os.path.exists(source_file):            

        # form aws cp command for this file
        aws_cmd = "aws s3 cp %s %s"%( source_file, destination_s3_url)

        subprocess.check_output( aws_cmd , stderr=subprocess.STDOUT, shell=True )

    else:
        raise Exception("Source file does not exist")

def load_image(image_source, image_name, docker_client):

    # if image source is docker hub, pull down image for this container
    # o.w. assume image is local
    if image_source == "hub":
        docker_client.pull( image_name )

# Load files specified in start configs
# Only s3 destinations are currently supported.
def load_files(files, template_vars):

    for file_key in files:

        this_file = files[file_key]

        # if destination is s3 bucket
        if is_s3_destination(this_file['dest']):

            # replace any template variables in the file before storing to S3
            with open(this_file['src'], 'r') as source_file:
                template = Template(source_file.read())

            # create a 'tmp' directory to hold the rendered source file
            if not os.path.exists('tmp'):
                os.makedirs('tmp')

            rendered_file = os.path.join('tmp','rendered_file')

            print 'rendered file: %s'%rendered_file
            with open(rendered_file, 'w') as outfile:
                outfile.write(template.render(template_vars))                

            # replace any template variables in the destination
            destination =  _render_string_variables(this_file['dest'],template_vars)

            cp_to_s3( rendered_file, destination)

        else:

            print '%s load was not performed. Only s3 destinations are currently supported.'%this_file['src']

# returns true if file to be loaded is configured for an s3 destination
def is_s3_destination( destination ):

    s3_destination = False

    # S3 destinations are URL's with s3 as the scheme
    # Use this to detect an S3 destination

    # attempt to parse the destination as a URL
    url_parts = urlparse.urlparse(destination)

    if (url_parts and url_parts.scheme and url_parts.scheme == 's3'):

        s3_destination = True

    return s3_destination


# converts a volumes_from list of container aliases to a list with container id's
def volumnes_from_with_ids(volumes_from, connection_str):

    volumes_from_ids = []

    for container_alias in volumes_from:

        # get the container
        container = find_container_by_name(container_alias, connection_str)

        volumes_from_ids.append(container.get('Id'))

    return volumes_from_ids


# start a container from an image
def start(
        image_tag,
        envs,
        alias,
        volumes_bindings,
        volume_mapping,
        volumes_from,
        port_bindings,
        connection_str,
        start_cmd,
        files_to_load,        
        template_vars,
        image_source='local',
        create_only=False):

    docker_client = get_docker_client( connection_str )

    # create the volumne host config from the volume_mapping
    volume_host_config = docker_client.create_host_config(binds=volume_mapping)

    # make sure container doesn't already exist
    existing_container = find_container_by_name( alias, connection_str )

    if existing_container:

        # stop container gracefully (giving it a full 30 secs)
        docker_client.stop(container=existing_container.get('Id'),
                           timeout=30)

        # remove container
        docker_client.remove_container( existing_container, force=True)

    # load the image on the host
    load_image(image_source, image_tag, docker_client)

    # load any files specified
    if files_to_load:

        load_files(files_to_load, template_vars)

    # render templates in environment variables
    rendered_envs = _render_env_variables(envs, template_vars)

    # create new container

    # if starting with a command other than the stock CMD in the dockerfile ...
    if start_cmd:

        container = docker_client.create_container(image=image_tag, 
                                                   name=alias, 
                                                   environment=rendered_envs,
                                                   volumes=volumes_bindings,                                               
                                                   host_config=volume_host_config,
                                                   command=start_cmd,
                                                   stdin_open=True,
                                                   tty=True)

    else:
        container = docker_client.create_container(image=image_tag, 
                                                   name=alias, 
                                                   environment=rendered_envs,
                                                   volumes=volumes_bindings,                                               
                                                   host_config=volume_host_config,
                                                   stdin_open=True,
                                                   tty=True)
    # start container
    if (container and container.get("Id") and not create_only):

        response = docker_client.start(container=container.get('Id'), 
                            port_bindings=port_bindings,
                            volumes_from = volumnes_from_with_ids(volumes_from,connection_str)  )

        if response:
            print "%s container created and started. Response from Docker: %s"%(alias,response)
        else:
            print "%s container created and started."%(alias)

    else:
        print "%s container was not started successfully"%alias



#!/usr/bin/env python
import time, os, json
from yac.lib.variables import get_variable, set_variable

# default yac namer

def get_stack_name( params ):  
    
    delimitter = get_variable(params,'delimitter','-')
    
    name_parts = [get_variable(params,'prefix',''), 
                    get_variable(params,'service-alias',''),
                    get_variable(params,'env','') ]

    # get rid of empty strings
    name_parts = filter(None,name_parts)

    stack_name = delimitter.join(name_parts)

    return stack_name

# name each yac resource
def get_resource_name(params, resource):   

    delimitter = get_variable(params,'delimitter','-')

    name_parts = [get_variable(params,'service-alias',''),
                  get_variable(params,'env',''),
                  resource]

    # get rid of empty strings
    name_parts = filter(None,name_parts)

    resource_name = delimitter.join(name_parts)

    return resource_name  


#!/usr/bin/env python

import os, argparse
from yac.lib.stack import get_stack_name, get_ec2_ips
from yac.lib.vpc import get_vpc_prefs

def is_valid_file(parser, arg):
    if not os.path.exists(arg):
        parser.error("The file %s does not exist!" % arg)
    else:
        return arg

def main():

    parser = argparse.ArgumentParser('ssh into a stack ec2 instance')

    parser.add_argument('app',  help='the app alias')
    parser.add_argument('env',  help='the env', choices=['dev', 'stage', 'prod', 'archive'])
    parser.add_argument('path', help='path to the private key file for this app',
                                type=lambda x: is_valid_file(parser, x))
    parser.add_argument('-s','--search',  help='search string, to support multiple EC2 instances in the same auto-scaling group')

    # pull out args
    args = parser.parse_args()

    # get any vpc preferences in place
    vpc_prefs = get_vpc_prefs()

    vpc_params = {}
    if 'vpc-params' in vpc_prefs:
        vpc_params = vpc_prefs['vpc-params']

    vpc_params.update({"service-alias": {"value": args.app}, "env": {"value": args.env}} )

    # get the name of the stack that corresponds to this app and env
    stack_name = get_stack_name(vpc_params)

    # get the IP address of the ec2 instance for this stack
    ec2_ips = get_ec2_ips(stack_name, args.search)

    # get the user to ssh in as
    if args.app != 'hipchat':
        ssh_user='core'
    else:
        ssh_user='admin'

    if (ec2_ips and len(ec2_ips)==1):
        ec2_ip = ec2_ips[0]

        # prepare ssh command
        ssh_cmd = "ssh -i %s %s@%s"%(args.path,ssh_user,ec2_ip)

        # ssh in
        os.system(ssh_cmd)

        print "future versions will remove 22 ingress here..."

    elif (ec2_ips and len(ec2_ips)>1):

        print 'Found multiple EC2 address for the %s app in the %s env. Consider adding a search string.'%(args.app, args,env)

    else:

        print 'Could not find address for the %s app in the %s env'%(args.app, args,env)

